# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta
from copy import deepcopy
from typing import Dict, List, Optional, Union

import torch
import torch.nn as nn
from mmengine import Config, MessageHub
from mmengine.model import BaseModel, is_model_wrapper
from mmengine.optim import OptimWrapper, OptimWrapperDict
from torch import Tensor

from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import ForwardInputs, NoiseVar, SampleList
from ..utils import (get_valid_noise_size, get_valid_num_batches,
                     noise_sample_fn, set_requires_grad)

ModelType = Union[Dict, nn.Module]


class BaseGAN(BaseModel, metaclass=ABCMeta):
    """Base class for GAN models.

    Args:
        generator (ModelType): The config or model of the generator.
        discriminator (Optional[ModelType]): The config or model of the
            discriminator. Defaults to None.
        data_preprocessor (Optional[Union[dict, Config]]): The pre-process
            config or :class:`~mmagic.models.DataPreprocessor`.
        generator_steps (int): The number of times the generator is completely
            updated before the discriminator is updated. Defaults to 1.
        discriminator_steps (int): The number of times the discriminator is
            completely updated before the generator is updated. Defaults to 1.
        ema_config (Optional[Dict]): The config for generator's exponential
            moving average setting. Defaults to None.
    """

    def __init__(self,
                 generator: ModelType,
                 discriminator: Optional[ModelType] = None,
                 data_preprocessor: Optional[Union[dict, Config]] = None,
                 generator_steps: int = 1,
                 discriminator_steps: int = 1,
                 noise_size: Optional[int] = None,
                 ema_config: Optional[Dict] = None,
                 loss_config: Optional[Dict] = None):
        super().__init__(data_preprocessor=data_preprocessor)

        # get valid noise_size
        noise_size = get_valid_noise_size(noise_size, generator)

        # build generator
        if isinstance(generator, dict):
            self._gen_cfg = deepcopy(generator)
            # build generator with default `noise_size` and `num_classes`
            gen_args = dict()
            if noise_size:
                gen_args['noise_size'] = noise_size
            if hasattr(self, 'num_classes') and self.num_classes is not None:
                gen_args['num_classes'] = self.num_classes
            generator = MODELS.build(generator, default_args=gen_args)
        self.generator = generator
        # get noise_size from generator because generator may have default
        # `noise_size` value
        self.noise_size = getattr(self.generator, 'noise_size', noise_size)

        # build discriminator
        if discriminator:
            if isinstance(discriminator, dict):
                self._disc_cfg = deepcopy(discriminator)
                # build discriminator with default `num_classes`
                disc_args = dict()
                if hasattr(self, 'num_classes'):
                    disc_args['num_classes'] = self.num_classes
                discriminator = MODELS.build(
                    discriminator, default_args=disc_args)
        self.discriminator = discriminator

        self._gen_steps = generator_steps
        self._disc_steps = discriminator_steps

        if ema_config is None:
            self._ema_config = None
            self._with_ema_gen = False
        else:
            self._ema_config = deepcopy(ema_config)
            self._init_ema_model(self._ema_config)
            self._with_ema_gen = True

        self._init_loss(loss_config)

    @staticmethod
    def gather_log_vars(log_vars_list: List[Dict[str, Tensor]]
                        ) -> Dict[str, Tensor]:
        """Gather a list of log_vars.
        Args:
            log_vars_list: List[Dict[str, Tensor]]

        Returns:
            Dict[str, Tensor]
        """
        if len(log_vars_list) == 1:
            return log_vars_list[0]

        log_keys = log_vars_list[0].keys()

        log_vars = dict()
        for k in log_keys:
            assert all([k in log_vars for log_vars in log_vars_list
                        ]), (f'\'{k}\' not in some of the \'log_vars\'.')
            log_vars[k] = torch.mean(
                torch.stack([log_vars[k] for log_vars in log_vars_list],
                            dim=0))

        return log_vars

    def _init_loss(self, loss_config: Optional[Dict] = None) -> None:
        """Initialize customized loss modules.

        If loss_config is a dict, we allow kinds of value for each field.

        1. `loss_config` is None: Users will implement all loss calculations
            in their own function. Weights for each loss terms are hard coded.
        2. `loss_config` is dict of scalar or string: Users will implement all
            loss calculations and use passed `loss_config` to control the
            weight or behavior of the loss calculation. Users will unpack and
            use each field in this dict by themselves.

            loss_config = dict(gp_norm_mode='HWC', gp_loss_weight=10)

        3. `loss_config` is dict of dict: Each field in `loss_config` will
            used to build a corresponding loss module. And use loss calculation
            function predefined by :class:`BaseGAN` to calculate the loss.

            loss_config = dict()

        Example:
            loss_config = dict(
                # `BaseGAN` pre-defined fields
                gan_loss=dict(type='GANLoss', gan_type='wgan-logistic-ns'),
                disc_auxiliary_loss=dict(
                    type='R1GradientPenalty',
                    loss_weight=10. / 2.,
                    interval=2,
                    norm_mode='HWC',
                    data_info=dict(
                        real_data='real_imgs',
                        discriminator='disc')),
                gen_auxiliary_loss=dict(
                    type='GeneratorPathRegularizer',
                    loss_weight=2,
                    pl_batch_shrink=2,
                    interval=g_reg_interval,
                    data_info=dict(
                        generator='gen',
                        num_batches='batch_size')),
                # user-defined field for loss weights or loss calculation
                my_loss_2=dict(weight=2, norm_mode='L1'),
                my_loss_3=2,
                my_loss_4_norm_type='L2')


        Args:
            loss_config (Optional[Dict], optional): Loss config used to build
                loss modules or define the loss weights. Defaults to None.
        """
        if loss_config is None:
            self.gan_loss = None
            self.gen_auxiliary_losses = None
            self.disc_auxiliary_losses = None
            self.loss_config = dict()
            return

        self.loss_config = deepcopy(loss_config)

        # build pre-defined losses
        gan_loss = loss_config.get('gan_loss', None)
        if gan_loss is not None:
            self.gan_loss = MODELS.build(gan_loss)
        else:
            self.gan_loss = None

        disc_auxiliary_loss = loss_config.get('disc_auxiliary_loss', None)
        if disc_auxiliary_loss:
            if not isinstance(disc_auxiliary_loss, list):
                disc_auxiliary_loss = [disc_auxiliary_loss]
            self.disc_auxiliary_losses = nn.ModuleList(
                [MODELS.build(loss) for loss in disc_auxiliary_loss])
        else:
            self.disc_auxiliary_losses = None

        gen_auxiliary_loss = loss_config.get('gen_auxiliary_loss', None)
        if gen_auxiliary_loss:
            if not isinstance(gen_auxiliary_loss, list):
                gen_auxiliary_loss = [gen_auxiliary_loss]
            self.gen_auxiliary_losses = nn.ModuleList(
                [MODELS.build(loss) for loss in gen_auxiliary_loss])
        else:
            self.gen_auxiliary_losses = None

    def noise_fn(self, noise: NoiseVar = None, num_batches: int = 1):
        """Sampling function for noise. There are three scenarios in this
        function:

        - If `noise` is a callable function, sample `num_batches` of noise
          with passed `noise`.
        - If `noise` is `None`, sample `num_batches` of noise from gaussian
          distribution.
        - If `noise` is a `torch.Tensor`, directly return `noise`.

        Args:
            noise (Union[Tensor, Callable, List[int], None]): You can directly
                give a batch of label through a ``torch.Tensor`` or offer a
                callable function to sample a batch of label data. Otherwise,
                the ``None`` indicates to use the default noise sampler.
                Defaults to `None`.
            num_batches (int, optional): The number of batches label want to
                sample. If `label` is a Tensor, this will be ignored. Defaults
                to 1.

        Returns:
            Tensor: Sampled noise tensor.
        """
        return noise_sample_fn(
            noise=noise,
            num_batches=num_batches,
            noise_size=self.noise_size,
            device=self.device)

    @property
    def generator_steps(self) -> int:
        """int: The number of times the generator is completely updated before
        the discriminator is updated."""
        return self._gen_steps

    @property
    def discriminator_steps(self) -> int:
        """int: The number of times the discriminator is completely updated
        before the generator is updated."""
        return self._disc_steps

    @property
    def device(self) -> torch.device:
        """Get current device of the model.

        Returns:
            torch.device: The current device of the model.
        """
        return next(self.parameters()).device

    @property
    def with_ema_gen(self) -> bool:
        """Whether the GAN adopts exponential moving average.

        Returns:
            bool: If `True`, means this GAN model is adopted to exponential
                moving average and vice versa.
        """
        return self._with_ema_gen

    def _init_ema_model(self, ema_config: dict):
        """Initialize a EMA model corresponding to the given `ema_config`. If
        `ema_config` is an empty dict or `None`, EMA model will not be
        initialized.

        Args:
            ema_config (dict): Config to initialize the EMA model.
        """
        ema_config.setdefault('type', 'ExponentialMovingAverage')
        self.ema_start = ema_config.pop('start_iter', 0)
        src_model = self.generator.module if is_model_wrapper(
            self.generator) else self.generator
        self.generator_ema = MODELS.build(
            ema_config, default_args=dict(model=src_model))

    def _get_valid_model(self, batch_inputs: ForwardInputs) -> str:
        """Try to get the valid forward model from inputs.

        - If forward model is defined in `batch_inputs`, it will be used as
          forward model.
        - If forward model is not defined in `batch_inputs`, 'ema' will
          returned if :property:`with_ema_gen` is true. Otherwise, 'orig' will
          be returned.

        Args:
            batch_inputs (ForwardInputs): Inputs passed to :meth:`forward`.

        Returns:
            str: Forward model to generate image. ('orig', 'ema' or
                'ema/orig').
        """
        if isinstance(batch_inputs, dict):
            sample_model = batch_inputs.get('sample_model', None)
        else:  # batch_inputs is a Tensor
            sample_model = None

        # set default value
        if sample_model is None:
            if self.with_ema_gen:
                sample_model = 'ema'
            else:
                sample_model = 'orig'

        # security checking for mode
        assert sample_model in [
            'ema', 'ema/orig', 'orig'
        ], ('Only support \'ema\', \'ema/orig\', \'orig\' '
            f'in {self.__class__.__name__}\'s image sampling.')
        if sample_model in ['ema', 'ema/orig']:
            assert self.with_ema_gen, (
                f'\'{self.__class__.__name__}\' do not have EMA model.')
        return sample_model

    def forward(self,
                inputs: ForwardInputs,
                data_samples: Optional[list] = None,
                mode: Optional[str] = None) -> SampleList:
        """Sample images with the given inputs. If forward mode is 'ema' or
        'orig', the image generated by corresponding generator will be
        returned. If forward mode is 'ema/orig', images generated by original
        generator and EMA generator will both be returned in a dict.

        Args:
            batch_inputs (ForwardInputs): Dict containing the necessary
                information (e.g. noise, num_batches, mode) to generate image.
            data_samples (Optional[list]): Data samples collated by
                :attr:`data_preprocessor`. Defaults to None.
            mode (Optional[str]): `mode` is not used in :class:`BaseGAN`.
                Defaults to None.

        Returns:
            SampleList: A list of ``DataSample`` contain generated results.
        """
        if isinstance(inputs, Tensor):
            noise = inputs
            sample_kwargs = {}
        else:
            noise = inputs.get('noise', None)
            num_batches = get_valid_num_batches(inputs, data_samples)
            noise = self.noise_fn(noise, num_batches=num_batches)
            sample_kwargs = inputs.get('sample_kwargs', dict())
        num_batches = noise.shape[0]

        sample_model = self._get_valid_model(inputs)
        batch_sample_list = []
        if sample_model in ['ema', 'orig']:
            if sample_model == 'ema':
                generator = self.generator_ema
            else:
                generator = self.generator
            if sample_kwargs:
                if 'return_noise' in sample_kwargs.keys():
                    outputs = generator(noise, **sample_kwargs)
                else:
                    outputs = generator(
                        noise, return_noise=False,
                        **sample_kwargs)  # yapf: disable
            else:
                sample_kwargs = {}
                outputs = generator(
                    noise, return_noise=False,
                    **sample_kwargs)  # no need to be False all time
            if isinstance(outputs, dict):
                if 'latent' in outputs.keys():
                    latent = outputs['latent']
                if 'feats' in outputs.keys():
                    feats = outputs['feats']
                outputs = outputs['fake_img']

            outputs = self.data_preprocessor.destruct(outputs, data_samples)

            gen_sample = DataSample()
            if data_samples:
                gen_sample.update(data_samples)
            if isinstance(inputs, dict) and 'img' in inputs:
                gen_sample.gt_img = inputs['img']
            gen_sample.fake_img = outputs
            gen_sample.noise = noise
            if 'latent' in locals():
                gen_sample.latent = latent
            if 'feats' in locals():
                gen_sample.feats = feats
            gen_sample.sample_kwargs = deepcopy(sample_kwargs)
            gen_sample.sample_model = sample_model
            batch_sample_list = gen_sample.split(allow_nonseq_value=True, )

        else:  # sample model is 'ema/orig
            outputs_orig = self.generator(
                noise, return_noise=False, **sample_kwargs)
            outputs_ema = self.generator_ema(
                noise, return_noise=False, **sample_kwargs)
            outputs_orig = self.data_preprocessor.destruct(
                outputs_orig, data_samples)
            outputs_ema = self.data_preprocessor.destruct(
                outputs_ema, data_samples)

            gen_sample = DataSample()
            if data_samples:
                gen_sample.update(data_samples)
            if isinstance(inputs, dict) and 'img' in inputs:
                gen_sample.gt_img = inputs['img']
            gen_sample.ema = DataSample(fake_img=outputs_ema)
            gen_sample.orig = DataSample(fake_img=outputs_orig)
            gen_sample.noise = noise
            gen_sample.sample_kwargs = deepcopy(sample_kwargs)
            gen_sample.sample_model = 'ema/orig'
            batch_sample_list = gen_sample.split(allow_nonseq_value=True)

        return batch_sample_list

    def val_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data.

        Calls ``self.data_preprocessor(data)`` and
        ``self(inputs, data_sample, mode=None)`` in order. Return the
        generated results which will be passed to evaluator.

        Args:
            data (dict): Data sampled from metric specific
                sampler. More details in `Metrics` and `Evaluator`.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        outputs = self(**data)
        return outputs

    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Same as :meth:`val_step`.

        Args:
            data (dict): Data sampled from metric specific
                sampler. More details in `Metrics` and `Evaluator`.

        Returns:
            List[DataSample]: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        outputs = self(**data)
        return outputs

    def train_step(self, data: dict,
                   optim_wrapper: OptimWrapperDict) -> Dict[str, Tensor]:
        """Train GAN model. In the training of GAN models, generator and
        discriminator are updated alternatively. In MMagic's design,
        `self.train_step` is called with data input. Therefore we always update
        discriminator, whose updating is relay on real data, and then determine
        if the generator needs to be updated based on the current number of
        iterations. More details about whether to update generator can be found
        in :meth:`should_gen_update`.

        Args:
            data (dict): Data sampled from dataloader.
            optim_wrapper (OptimWrapperDict): OptimWrapperDict instance
                contains OptimWrapper of generator and discriminator.

        Returns:
            Dict[str, torch.Tensor]: A ``dict`` of tensor for logging.
        """
        message_hub = MessageHub.get_current_instance()
        curr_iter = message_hub.get_info('iter')
        data = self.data_preprocessor(data, True)

        disc_optimizer_wrapper: OptimWrapper = optim_wrapper['discriminator']
        disc_accu_iters = disc_optimizer_wrapper._accumulative_counts

        with disc_optimizer_wrapper.optim_context(self.discriminator):
            log_vars = self.train_discriminator(
                **data, optimizer_wrapper=disc_optimizer_wrapper)

        # add 1 to `curr_iter` because iter is updated in train loop.
        # Whether to update the generator. We update generator with
        # discriminator is fully updated for `self.n_discriminator_steps`
        # iterations. And one full updating for discriminator contains
        # `disc_accu_counts` times of grad accumulations.
        if (curr_iter + 1) % (self.discriminator_steps * disc_accu_iters) == 0:
            set_requires_grad(self.discriminator, False)
            gen_optimizer_wrapper = optim_wrapper['generator']
            gen_accu_iters = gen_optimizer_wrapper._accumulative_counts

            log_vars_gen_list = []
            # init optimizer wrapper status for generator manually
            gen_optimizer_wrapper.initialize_count_status(
                self.generator, 0, self.generator_steps * gen_accu_iters)
            for _ in range(self.generator_steps * gen_accu_iters):
                with gen_optimizer_wrapper.optim_context(self.generator):
                    log_vars_gen = self.train_generator(
                        **data, optimizer_wrapper=gen_optimizer_wrapper)

                log_vars_gen_list.append(log_vars_gen)
            log_vars_gen = self.gather_log_vars(log_vars_gen_list)
            log_vars_gen.pop('loss', None)  # remove 'loss' from gen logs

            set_requires_grad(self.discriminator, True)

            # only do ema after generator update
            if self.with_ema_gen and (curr_iter + 1) >= (
                    self.ema_start * self.discriminator_steps *
                    disc_accu_iters):
                self.generator_ema.update_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)
                # if not update buffer, copy buffer from orig model
                if not self.generator_ema.update_buffers:
                    self.generator_ema.sync_buffers(
                        self.generator.module if is_model_wrapper(
                            self.generator) else self.generator)
            elif self.with_ema_gen:
                # before ema, copy weights from orig
                self.generator_ema.sync_parameters(
                    self.generator.module
                    if is_model_wrapper(self.generator) else self.generator)

            log_vars.update(log_vars_gen)

        return log_vars

    def _get_gen_loss(self, out_dict):
        losses_dict = {}
        # gan loss
        losses_dict['loss_disc_fake_g'] = self.gan_loss(
            out_dict['disc_pred_fake_g'], target_is_real=True, is_disc=False)

        # gen auxiliary loss
        if self.gen_auxiliary_losses is not None:
            for loss_module in self.gen_auxiliary_losses:
                loss_ = loss_module(out_dict)
                if loss_ is None:
                    continue

                # the `loss_name()` function return name as 'loss_xxx'
                if loss_module.loss_name() in losses_dict:
                    losses_dict[loss_module.loss_name(
                    )] = losses_dict[loss_module.loss_name()] + loss_
                else:
                    losses_dict[loss_module.loss_name()] = loss_
        loss, log_var = self.parse_losses(losses_dict)

        return loss, log_var

    def _get_disc_loss(self, out_dict):
        # Construct losses dict. If you hope some items to be included in the
        # computational graph, you have to add 'loss' in its name. Otherwise,
        # items without 'loss' in their name will just be used to print
        # information.
        losses_dict = {}
        # gan loss
        losses_dict['loss_disc_fake'] = self.gan_loss(
            out_dict['disc_pred_fake'], target_is_real=False, is_disc=True)
        losses_dict['loss_disc_real'] = self.gan_loss(
            out_dict['disc_pred_real'], target_is_real=True, is_disc=True)

        # disc auxiliary loss
        if self.disc_auxiliary_losses is not None:
            for loss_module in self.disc_auxiliary_losses:
                loss_ = loss_module(out_dict)
                if loss_ is None:
                    continue

                # the `loss_name()` function return name as 'loss_xxx'
                if loss_module.loss_name() in losses_dict:
                    losses_dict[loss_module.loss_name(
                    )] = losses_dict[loss_module.loss_name()] + loss_
                else:
                    losses_dict[loss_module.loss_name()] = loss_
        loss, log_var = self.parse_losses(losses_dict)

        return loss, log_var

    def train_generator(self, inputs: dict, data_samples: List[DataSample],
                        optimizer_wrapper: OptimWrapper) -> Dict[str, Tensor]:
        """Training function for discriminator. All GANs should implement this
        function by themselves.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (List[DataSample]): Data samples from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        num_batches = inputs['img'].shape[0]
        noise = self.noise_fn(num_batches=num_batches)

        fake_imgs = self.generator(noise=noise)
        disc_pred_fake_g = self.discriminator(fake_imgs)

        data_dict_ = dict(
            gen=self.generator,
            disc=self.discriminator,
            fake_imgs=fake_imgs,
            disc_pred_fake_g=disc_pred_fake_g,
            # iteration=curr_iter,
            batch_size=num_batches,
            loss_scaler=getattr(optimizer_wrapper, 'loss_scaler', None))
        loss, log_vars = self._get_gen_loss(data_dict_)

        optimizer_wrapper.update_params(loss)
        return log_vars

    def train_discriminator(self, inputs: dict, data_samples: List[DataSample],
                            optimizer_wrapper: OptimWrapper
                            ) -> Dict[str, Tensor]:
        """Training function for discriminator. All GANs should implement this
        function by themselves.

        Args:
            inputs (dict): Inputs from dataloader.
            data_samples (List[DataSample]): Data samples from dataloader.
            optim_wrapper (OptimWrapper): OptimWrapper instance used to update
                model parameters.

        Returns:
            Dict[str, Tensor]: A ``dict`` of tensor for logging.
        """
        real_imgs, num_batches = inputs['img'], inputs['img'].shape[0]
        noise = self.noise_fn(num_batches=num_batches)
        fake_imgs = self.generator(noise=noise)

        # disc pred for fake imgs and real_imgs
        disc_pred_fake = self.discriminator(fake_imgs)
        disc_pred_real = self.discriminator(real_imgs)
        # get data dict to compute losses for disc
        data_dict_ = dict(
            gen=self.generator,
            disc=self.discriminator,
            disc_pred_fake=disc_pred_fake,
            disc_pred_real=disc_pred_real,
            fake_imgs=fake_imgs,
            real_imgs=real_imgs,
            # iteration=curr_iter,
            batch_size=num_batches,
            loss_scaler=getattr(optimizer_wrapper, 'loss_scaler', None))
        loss, log_vars = self._get_disc_loss(data_dict_)

        optimizer_wrapper.update_params(loss)
        return log_vars



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
from argparse import ArgumentParser
from functools import partial

import gradio as gr
import numpy as np
import torch
from PIL import Image

from demo.utils.gradio_utils import (EasyDict, ImageMask, draw_mask_on_image,
                                     draw_points_on_image,
                                     get_latest_points_pair, get_valid_mask,
                                     on_change_single_global_state)
from demo.utils.renderer import Renderer, add_watermark_np

parser = ArgumentParser()
parser.add_argument('--share', action='store_true', default='True')
parser.add_argument('--cache-dir', type=str, default='./checkpoints')
parser.add_argument(
    '--listen',
    action='store_true',
    help=''
    'launch gradio with 0.0.0.0 as server name, \
        allowing to respond to network requests',
)
args = parser.parse_args()
cache_dir = args.cache_dir


def get_default_gr_state(device='cuda',
                         init_pkl='stylegan2_lions_512_pytorch_mmagic.pth'):
    global_state = gr.State({
        'images': {
            # image_orig: the original image, change with seed/model is changed
            # image_raw: image with mask and points, change during optimization
            # image_show: image showed on screen
        },
        'temporal_params': {
            # stop
        },
        'mask':
        None,  # mask for visualization, 1 for editing and 0 for unchange
        'last_mask': None,  # last edited mask
        'show_mask': True,  # add button
        'generator_params': EasyDict(),
        'params': {
            'seed': 0,
            'motion_lambda': 20,
            'r1_in_pixels': 3,
            'r2_in_pixels': 12,
            'magnitude_direction_in_pixels': 1.0,
            'latent_space': 'w+',
            'trunc_psi': 0.7,
            'trunc_cutoff': None,
            'lr': 0.001,
        },
        'device': device,
        'draw_interval': 1,
        'renderer': Renderer(disable_timing=True),
        'points': {},
        'curr_point': None,
        'curr_type_point': 'start',
        'editing_state': 'add_points',
        'pretrained_weight': init_pkl
    })
    return global_state


def reverse_point_pairs(points):
    new_points = []
    for p in points:
        new_points.append([p[1], p[0]])
    return new_points


def clear_state(global_state, target=None):
    """Clear target history state from global_state If target is not defined,
    points and mask will be both removed.

    1. set global_state['points'] as empty dict
    2. set global_state['mask'] as full-one mask.
    """
    if target is None:
        target = ['point', 'mask']
    if not isinstance(target, list):
        target = [target]
    if 'point' in target:
        global_state['points'] = dict()
        print('Clear Points State!')
    if 'mask' in target:
        image_raw = global_state['images']['image_raw']
        global_state['mask'] = np.ones((image_raw.size[1], image_raw.size[0]),
                                       dtype=np.uint8)
        print('Clear mask State!')

    return global_state


def init_images(global_state, cache_dir='./checkpoints'):
    """This function is called only ones with Gradio App is started.

    0. pre-process global_state, unpack value from global_state of need
    1. Re-init renderer
    2. run `renderer._render_drag_impl` with `is_drag=False` to generate
       new image
    3. Assign images to global state and re-generate mask
    """

    if isinstance(global_state, gr.State):
        state = global_state.value
    else:
        state = global_state

    state['renderer'].init_network(
        state['generator_params'],  # res
        os.path.join(cache_dir, state['pretrained_weight']
                     ),  # StyleGAN2 checkpoint path from Official DragGAN
        state['params']['seed'],  # w0_seed,
        None,  # w_load
        state['params']['latent_space'] == 'w+',  # w_plus
        'const',
        state['params']['trunc_psi'],  # trunc_psi,
        state['params']['trunc_cutoff'],  # trunc_cutoff,
        None,  # input_transform
        state['params']['lr']  # lr,
    )
    state['renderer']._render_drag_impl(
        state['generator_params'], is_drag=False, to_pil=True)

    init_image = state['generator_params'].image
    state['images']['image_orig'] = init_image
    state['images']['image_raw'] = init_image
    state['images']['image_show'] = Image.fromarray(
        add_watermark_np(np.array(init_image)))
    state['mask'] = np.ones((init_image.size[1], init_image.size[0]),
                            dtype=np.uint8)
    return global_state


def update_image_draw(image, points, mask, show_mask, global_state=None):
    image_draw = draw_points_on_image(image, points)
    if show_mask and mask is not None and not (mask == 0).all() and not (
            mask == 1).all():
        image_draw = draw_mask_on_image(image_draw, mask)

    image_draw = Image.fromarray(add_watermark_np(np.array(image_draw)))
    if global_state is not None:
        global_state['images']['image_show'] = image_draw
    return image_draw


def preprocess_mask_info(global_state, image):
    """Function to handle mask information.

    1. last_mask is None: Do not need to change mask, return mask
    2. last_mask is not None:
        2.1 global_state is remove_mask:
        2.2 global_state is add_mask:
    """

    if isinstance(image, dict):
        last_mask = get_valid_mask(image['mask'])
    else:
        last_mask = None
    mask = global_state['mask']
    # mask in global state is a placeholder with all 1.
    if (mask == 1).all():
        mask = last_mask

    # last_mask = global_state['last_mask']
    editing_mode = global_state['editing_state']

    if last_mask is None:
        return global_state

    if editing_mode == 'remove_mask':
        updated_mask = np.clip(mask - last_mask, 0, 1)
        print(f'Last editing_state is {editing_mode}, do remove.')
    elif editing_mode == 'add_mask':
        updated_mask = np.clip(mask + last_mask, 0, 1)
        print(f'Last editing_state is {editing_mode}, do add.')
    else:
        updated_mask = mask
        print(f'Last editing_state is {editing_mode}, ' 'do nothing to mask.')

    global_state['mask'] = updated_mask
    # global_state['last_mask'] = None  # clear buffer
    return global_state


valid_checkpoints_dict = {
    f.split('/')[-1]: osp.join(cache_dir, f)
    for f in os.listdir(cache_dir)
    if (f.endswith('pth') and osp.exists(osp.join(cache_dir, f)))
}
print(f'File under cache_dir ({cache_dir}):')
print(os.listdir(cache_dir))
print('Valid checkpoint file:')
print(valid_checkpoints_dict)

if __name__ == '__main__':
    init_pkl = 'stylegan2_lions_512_pytorch_mmagic.pth'
    device = 'cuda'

    with gr.Blocks() as app:

        # renderer = Renderer()
        global_state = get_default_gr_state(device, init_pkl)

        # init image
        global_state = init_images(global_state)

        with gr.Row():

            with gr.Row():

                # Left --> tools
                with gr.Column(scale=3):

                    # Pickle
                    with gr.Row():

                        with gr.Column(scale=1, min_width=10):
                            gr.Markdown(value='Pickle', show_label=False)

                        with gr.Column(scale=4, min_width=10):
                            form_pretrained_dropdown = gr.Dropdown(
                                choices=list(valid_checkpoints_dict.keys()),
                                label='Pretrained Model',
                                value=init_pkl,
                            )

                    # Latent
                    with gr.Row():
                        with gr.Column(scale=1, min_width=10):
                            gr.Markdown(value='Latent', show_label=False)

                        with gr.Column(scale=4, min_width=10):
                            form_seed_number = gr.Number(
                                value=global_state.value['params']['seed'],
                                interactive=True,
                                label='Seed',
                            )
                            form_lr_number = gr.Number(
                                value=global_state.value['params']['lr'],
                                interactive=True,
                                label='Step Size')

                            with gr.Row():
                                with gr.Column(scale=2, min_width=10):
                                    form_reset_image = gr.Button('Reset Image')
                                with gr.Column(scale=3, min_width=10):
                                    form_latent_space = gr.Radio(
                                        ['w', 'w+'],
                                        value=global_state.value['params']
                                        ['latent_space'],
                                        interactive=True,
                                        label='Latent space to optimize',
                                        show_label=False,
                                    )

                    # Drag
                    with gr.Row():
                        with gr.Column(scale=1, min_width=10):
                            gr.Markdown(value='Drag', show_label=False)
                        with gr.Column(scale=4, min_width=10):
                            with gr.Row():
                                with gr.Column(scale=1, min_width=10):
                                    enable_add_points = gr.Button('Add Points')
                                with gr.Column(scale=1, min_width=10):
                                    undo_points = gr.Button('Reset Points')
                            with gr.Row():
                                with gr.Column(scale=1, min_width=10):
                                    form_start_btn = gr.Button('Start')
                                with gr.Column(scale=1, min_width=10):
                                    form_stop_btn = gr.Button('Stop')

                            form_steps_number = gr.Number(
                                value=0, label='Steps', interactive=False)

                    # Mask
                    with gr.Row():
                        with gr.Column(scale=1, min_width=10):
                            gr.Markdown(value='Mask', show_label=False)
                        with gr.Column(scale=4, min_width=10):
                            enable_add_mask = gr.Button('Edit Flexible Area')
                            with gr.Row():
                                with gr.Column(scale=1, min_width=10):
                                    form_reset_mask_btn = gr.Button(
                                        'Reset mask')
                                with gr.Column(scale=1, min_width=10):
                                    show_mask = gr.Checkbox(
                                        label='Show Mask',
                                        value=global_state.value['show_mask'],
                                        show_label=False)

                            with gr.Row():
                                form_lambda_number = gr.Number(
                                    value=global_state.value['params']
                                    ['motion_lambda'],
                                    interactive=True,
                                    label='Lambda',
                                )

                    form_draw_interval_number = gr.Number(
                        value=global_state.value['draw_interval'],
                        label='Draw Interval (steps)',
                        interactive=True,
                        visible=False)

                # Right --> Image
                with gr.Column(scale=8):
                    form_image = ImageMask(
                        value=global_state.value['images']['image_show'],
                        brush_radius=20).style(
                            width=768,
                            height=768)  # NOTE: hard image size code here.
        gr.Markdown("""
            ## Quick Start

            1. Select desired `Pretrained Model` and adjust `Seed` to generate
            an initial image.
            2. Click on image to add control points.
            3. Click `Start` and enjoy it!

            ## Advance Usage

            1. Change `Step Size` to adjust learning rate in drag optimization.
            2. Select `w` or `w+` to change latent space to optimize:
            * Optimize on `w` space may cause greater influence to the image.
            * Optimize on `w+` space may work slower than `w`, but usually
            achieve better results.
            * Note that changing the latent space will reset the image, points
            and mask (this has the same effect as `Reset Image` button).
            3. Click `Edit Flexible Area` to create a mask and constrain the
            unmasked region to remain unchanged.
            """)
        gr.HTML("""
            <style>
                .container {
                    position: absolute;
                    height: 50px;
                    text-align: center;
                    line-height: 50px;
                    width: 100%;
                }
            </style>
            <div class="container">
            Gradio demo supported by
            <img src="https://avatars.githubusercontent.com/u/10245193?s=
            200&v=4"\
                height="20" width="20" style="display:inline;">
            <a href="https://github.com/open-mmlab/mmagic">OpenMMLab MMagic</a>
            </div>
            """)

        # Network & latents tab listeners
        def on_change_pretrained_dropdown(pretrained_value, global_state):
            """Function to handle model change.

            1. Set pretrained value to global_state
            2. Re-init images and clear all states
            """

            global_state['pretrained_weight'] = pretrained_value
            init_images(global_state)
            clear_state(global_state)

            return global_state, global_state['images']['image_show']

        form_pretrained_dropdown.change(
            on_change_pretrained_dropdown,
            inputs=[form_pretrained_dropdown, global_state],
            outputs=[global_state, form_image],
        )

        def on_click_reset_image(global_state):
            """Reset image to the original one and clear all states
            1. Re-init images
            2. Clear all states
            """

            init_images(global_state)
            clear_state(global_state)

            return global_state, global_state['images']['image_show']

        form_reset_image.click(
            on_click_reset_image,
            inputs=[global_state],
            outputs=[global_state, form_image],
        )

        # Update parameters
        def on_change_update_image_seed(seed, global_state):
            """Function to handle generation seed change.

            1. Set seed to global_state
            2. Re-init images and clear all states
            """

            global_state['params']['seed'] = int(seed)
            init_images(global_state)
            clear_state(global_state)

            return global_state, global_state['images']['image_show']

        form_seed_number.change(
            on_change_update_image_seed,
            inputs=[form_seed_number, global_state],
            outputs=[global_state, form_image],
        )

        def on_click_latent_space(latent_space, global_state):
            """Function to reset latent space to optimize.

            NOTE: this function we reset the image and all controls
            1. Set latent-space to global_state
            2. Re-init images and clear all state
            """

            global_state['params']['latent_space'] = latent_space
            init_images(global_state)
            clear_state(global_state)

            return global_state, global_state['images']['image_show']

        form_latent_space.change(
            on_click_latent_space,
            inputs=[form_latent_space, global_state],
            outputs=[global_state, form_image])

        # ==== Params
        form_lambda_number.change(
            partial(on_change_single_global_state,
                    ['params', 'motion_lambda']),
            inputs=[form_lambda_number, global_state],
            outputs=[global_state],
        )

        def on_change_lr(lr, global_state):
            if lr == 0:
                print('lr is 0, do nothing.')
                return global_state
            else:
                global_state['params']['lr'] = lr
                renderer = global_state['renderer']
                renderer.update_lr(lr)
                print('New optimizer: ')
                print(renderer.w_optim)
            return global_state

        form_lr_number.change(
            on_change_lr,
            inputs=[form_lr_number, global_state],
            outputs=[global_state],
        )

        def on_click_start(global_state, image):
            p_in_pixels = []
            t_in_pixels = []
            valid_points = []

            # handle of start drag in mask editing mode
            global_state = preprocess_mask_info(global_state, image)

            # Prepare the points for the inference
            if len(global_state['points']) == 0:
                # yield on_click_start_wo_points(global_state, image)
                image_raw = global_state['images']['image_raw']
                update_image_draw(
                    image_raw,
                    global_state['points'],
                    global_state['mask'],
                    global_state['show_mask'],
                    global_state,
                )

                yield (
                    global_state,
                    0,
                    global_state['images']['image_show'],
                    # gr.File.update(visible=False),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    # latent space
                    gr.Radio.update(interactive=True),
                    gr.Button.update(interactive=True),
                    # NOTE: disable stop button
                    gr.Button.update(interactive=False),

                    # update other comps
                    gr.Dropdown.update(interactive=True),
                    gr.Number.update(interactive=True),
                    gr.Number.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Checkbox.update(interactive=True),
                    # gr.Number.update(interactive=True),
                    gr.Number.update(interactive=True),
                )
            else:

                # Transform the points into torch tensors
                for key_point, point in global_state['points'].items():
                    try:
                        p_start = point.get('start_temp', point['start'])
                        p_end = point['target']

                        if p_start is None or p_end is None:
                            continue

                    except KeyError:
                        continue

                    p_in_pixels.append(p_start)
                    t_in_pixels.append(p_end)
                    valid_points.append(key_point)

                mask = torch.tensor(global_state['mask']).float()
                drag_mask = 1 - mask

                renderer: Renderer = global_state['renderer']
                global_state['temporal_params']['stop'] = False
                global_state['editing_state'] = 'running'

                # reverse points order
                p_to_opt = reverse_point_pairs(p_in_pixels)
                t_to_opt = reverse_point_pairs(t_in_pixels)
                print(p_to_opt)
                print('Running with:')
                print(f'    Source: {p_in_pixels}')
                print(f'    Target: {t_in_pixels}')
                step_idx = 0
                while True:
                    if global_state['temporal_params']['stop']:
                        break

                    # do drage here!
                    renderer._render_drag_impl(
                        global_state['generator_params'],
                        p_to_opt,  # point
                        t_to_opt,  # target
                        drag_mask,  # mask,
                        global_state['params']['motion_lambda'],  # lambda_mask
                        reg=0,
                        feature_idx=5,  # NOTE: do not support change for now
                        r1=global_state['params']['r1_in_pixels'],  # r1
                        r2=global_state['params']['r2_in_pixels'],  # r2
                        # random_seed     = 0,
                        # noise_mode      = 'const',
                        trunc_psi=global_state['params']['trunc_psi'],
                        # force_fp32      = False,
                        # layer_name      = None,
                        # sel_channels    = 3,
                        # base_channel    = 0,
                        # img_scale_db    = 0,
                        # img_normalize   = False,
                        # untransform     = False,
                        is_drag=True,
                        to_pil=True)

                    if step_idx % global_state['draw_interval'] == 0:
                        print('Current Source:')
                        for key_point, p_i, t_i in zip(valid_points, p_to_opt,
                                                       t_to_opt):
                            global_state['points'][key_point]['start_temp'] = [
                                p_i[1],
                                p_i[0],
                            ]
                            global_state['points'][key_point]['target'] = [
                                t_i[1],
                                t_i[0],
                            ]
                            start_temp = global_state['points'][key_point][
                                'start_temp']
                            print(f'    {start_temp}')

                        image_result = global_state['generator_params'][
                            'image']
                        image_draw = update_image_draw(  # noqa
                            image_result,
                            global_state['points'],
                            global_state['mask'],
                            global_state['show_mask'],
                            global_state,
                        )
                        global_state['images']['image_raw'] = image_result

                    yield (
                        global_state,
                        step_idx,
                        global_state['images']['image_show'],
                        # gr.File.update(visible=False),
                        gr.Button.update(interactive=False),
                        gr.Button.update(interactive=False),
                        gr.Button.update(interactive=False),
                        gr.Button.update(interactive=False),
                        gr.Button.update(interactive=False),
                        # latent space
                        gr.Radio.update(interactive=False),
                        gr.Button.update(interactive=False),
                        # enable stop button in loop
                        gr.Button.update(interactive=True),

                        # update other comps
                        gr.Dropdown.update(interactive=False),
                        gr.Number.update(interactive=False),
                        gr.Number.update(interactive=False),
                        gr.Button.update(interactive=False),
                        gr.Button.update(interactive=False),
                        gr.Checkbox.update(interactive=False),
                        # gr.Number.update(interactive=False),
                        gr.Number.update(interactive=False),
                    )

                    # increate step
                    step_idx += 1

                image_result = global_state['generator_params']['image']
                global_state['images']['image_raw'] = image_result
                # image_draw = update_image_draw(image_result,
                #                                global_state['points'],
                #                                global_state['mask'],
                #                                global_state['show_mask'],
                #                                global_state)

                # fp = NamedTemporaryFile(suffix=".png", delete=False)
                # image_result.save(fp, "PNG")

                global_state['editing_state'] = 'add_points'

                yield (
                    global_state,
                    0,  # reset step to 0 after stop.
                    global_state['images']['image_show'],
                    # gr.File.update(visible=True, value=fp.name),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    gr.Button.update(interactive=True),
                    # latent space
                    gr.Radio.update(interactive=True),
                    gr.Button.update(interactive=True),
                    # NOTE: disable stop button with loop finish
                    gr.Button.update(interactive=False),

                    # update other comps
                    gr.Dropdown.update(interactive=True),
                    gr.Number.update(interactive=True),
                    gr.Number.update(interactive=True),
                    gr.Checkbox.update(interactive=True),
                    gr.Number.update(interactive=True),
                )

        form_start_btn.click(
            on_click_start,
            inputs=[global_state, form_image],
            outputs=[
                global_state,
                form_steps_number,
                form_image,
                # form_download_result_file,
                # >>> buttons
                form_reset_image,
                enable_add_points,
                enable_add_mask,
                undo_points,
                form_reset_mask_btn,
                form_latent_space,
                form_start_btn,
                form_stop_btn,
                # <<< buttonm
                # >>> inputs comps
                form_pretrained_dropdown,
                form_seed_number,
                form_lr_number,
                show_mask,
                form_lambda_number,
            ],
        )

        def on_click_stop(global_state):
            """Function to handle stop button is clicked.

            1. send a stop signal by
            set global_state["temporal_params"]["stop"] as True
            2. Disable Stop button
            """
            global_state['temporal_params']['stop'] = True

            return global_state, gr.Button.update(interactive=False)

        form_stop_btn.click(
            on_click_stop,
            inputs=[global_state],
            outputs=[global_state, form_stop_btn])

        form_draw_interval_number.change(
            partial(
                on_change_single_global_state,
                'draw_interval',
                map_transform=lambda x: int(x),
            ),
            inputs=[form_draw_interval_number, global_state],
            outputs=[global_state],
        )

        def on_click_remove_point(global_state):
            choice = global_state['curr_point']
            del global_state['points'][choice]

            choices = list(global_state['points'].keys())

            if len(choices) > 0:
                global_state['curr_point'] = choices[0]

            return (
                gr.Dropdown.update(choices=choices, value=choices[0]),
                global_state,
            )

        # Mask
        def on_click_reset_mask(global_state):
            global_state['mask'] = np.ones(
                (
                    global_state['images']['image_raw'].size[1],
                    global_state['images']['image_raw'].size[0],
                ),
                dtype=np.uint8,
            )
            image_draw = update_image_draw(global_state['images']['image_raw'],
                                           global_state['points'],
                                           global_state['mask'],
                                           global_state['show_mask'],
                                           global_state)
            return global_state, image_draw

        form_reset_mask_btn.click(
            on_click_reset_mask,
            inputs=[global_state],
            outputs=[global_state, form_image],
        )

        # Image
        def on_click_enable_draw(global_state, image):
            """Function to start add mask mode.

            1. Preprocess mask info from last state
            2. Change editing state to add_mask
            3. Set curr image with points and mask
            """
            global_state = preprocess_mask_info(global_state, image)
            global_state['editing_state'] = 'add_mask'
            image_raw = global_state['images']['image_raw']
            image_draw = update_image_draw(image_raw, global_state['points'],
                                           global_state['mask'], True,
                                           global_state)
            return (global_state,
                    gr.Image.update(value=image_draw, interactive=True))

        def on_click_remove_draw(global_state, image):
            """Function to start remove mask mode.

            1. Preprocess mask info from last state
            2. Change editing state to remove_mask
            3. Set curr image with points and mask
            """
            global_state = preprocess_mask_info(global_state, image)
            global_state['edinting_state'] = 'remove_mask'
            image_raw = global_state['images']['image_raw']
            image_draw = update_image_draw(image_raw, global_state['points'],
                                           global_state['mask'], True,
                                           global_state)
            return (global_state,
                    gr.Image.update(value=image_draw, interactive=True))

        enable_add_mask.click(
            on_click_enable_draw,
            inputs=[global_state, form_image],
            outputs=[
                global_state,
                form_image,
            ])

        def on_click_add_point(global_state, image: dict):
            """Function switch from add mask mode to add points mode.

            1. Updaste mask buffer if need
            2. Change global_state['editing_state'] to 'add_points'
            3. Set current image with mask
            """

            global_state = preprocess_mask_info(global_state, image)
            global_state['editing_state'] = 'add_points'
            mask = global_state['mask']
            image_raw = global_state['images']['image_raw']
            image_draw = update_image_draw(image_raw, global_state['points'],
                                           mask, global_state['show_mask'],
                                           global_state)

            return (global_state,
                    gr.Image.update(value=image_draw, interactive=False))

        enable_add_points.click(
            on_click_add_point,
            inputs=[global_state, form_image],
            outputs=[global_state, form_image])

        def on_click_image(global_state, evt: gr.SelectData):
            """This function only support click for point selection."""
            xy = evt.index
            if global_state['editing_state'] != 'add_points':
                print(f'In {global_state["editing_state"]} state. '
                      'Do not add points.')

                return global_state, global_state['images']['image_show']

            points = global_state['points']

            point_idx = get_latest_points_pair(points)
            if point_idx is None:
                points[0] = {'start': xy, 'target': None}
                print(f'Click Image - Start - {xy}')
            elif points[point_idx].get('target', None) is None:
                points[point_idx]['target'] = xy
                print(f'Click Image - Target - {xy}')
            else:
                points[point_idx + 1] = {'start': xy, 'target': None}
                print(f'Click Image - Start - {xy}')

            image_raw = global_state['images']['image_raw']
            image_draw = update_image_draw(
                image_raw,
                global_state['points'],
                global_state['mask'],
                global_state['show_mask'],
                global_state,
            )

            return global_state, image_draw

        form_image.select(
            on_click_image,
            inputs=[global_state],
            outputs=[global_state, form_image],
        )

        def on_click_clear_points(global_state):
            """Function to handle clear all control points
            1. clear global_state['points'] (clear_state)
            2. re-init network
            2. re-draw image
            """
            clear_state(global_state, target='point')

            renderer: Renderer = global_state['renderer']
            renderer.feat_refs = None

            image_raw = global_state['images']['image_raw']
            image_draw = update_image_draw(image_raw, {}, global_state['mask'],
                                           global_state['show_mask'],
                                           global_state)
            return global_state, image_draw

        undo_points.click(
            on_click_clear_points,
            inputs=[global_state],
            outputs=[global_state, form_image])

        def on_click_show_mask(global_state, show_mask):
            """Function to control whether show mask on image."""
            global_state['show_mask'] = show_mask

            image_raw = global_state['images']['image_raw']
            image_draw = update_image_draw(
                image_raw,
                global_state['points'],
                global_state['mask'],
                global_state['show_mask'],
                global_state,
            )
            return global_state, image_draw

        show_mask.change(
            on_click_show_mask,
            inputs=[global_state, show_mask],
            outputs=[global_state, form_image],
        )

    gr.close_all()
    app.queue(concurrency_count=3, max_size=20)
    app.launch(
        share=args.share,
        server_name='0.0.0.0' if args.listen else '127.0.0.1')



_base_ = [
    '../_base_/gen_default_runtime.py',
    '../_base_/models/base_styleganv2.py',
]

# reg params
d_reg_interval = 16
g_reg_interval = 4

g_reg_ratio = g_reg_interval / (g_reg_interval + 1)
d_reg_ratio = d_reg_interval / (d_reg_interval + 1)

ema_half_life = 10.  # G_smoothing_kimg

model = dict(
    generator=dict(out_size=512),
    discriminator=dict(in_size=512),
    ema_config=dict(
        type='ExponentialMovingAverage',
        interval=1,
        momentum=1. - (0.5**(32. / (ema_half_life * 1000.)))),
    loss_config=dict(
        r1_loss_weight=10. / 2. * d_reg_interval,
        r1_interval=d_reg_interval,
        norm_mode='HWC',
        g_reg_interval=g_reg_interval,
        g_reg_weight=2. * g_reg_interval,
        pl_batch_shrink=2))

train_cfg = dict(max_iters=1800002)

optim_wrapper = dict(
    generator=dict(
        optimizer=dict(
            type='Adam', lr=0.002 * g_reg_ratio, betas=(0,
                                                        0.99**g_reg_ratio))),
    discriminator=dict(
        optimizer=dict(
            type='Adam', lr=0.002 * d_reg_ratio, betas=(0,
                                                        0.99**d_reg_ratio))))
# DATA
batch_size = 4
data_root = './data/lsun/images/car'
dataset_type = 'BasicImageDataset'

train_pipeline = [
    dict(type='LoadImageFromFile', key='gt'),
    dict(
        type='NumpyPad',
        keys='img',
        padding=((64, 64), (0, 0), (0, 0)),
    ),
    dict(type='Flip', keys=['gt'], direction='horizontal'),
    dict(type='PackInputs')
]

val_pipeline = train_pipeline

# `batch_size` and `data_root` need to be set.
train_dataloader = dict(
    batch_size=4,
    num_workers=8,
    persistent_workers=True,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    dataset=dict(
        type=dataset_type, data_root=data_root, pipeline=train_pipeline))

val_dataloader = dict(
    batch_size=4,
    num_workers=8,
    dataset=dict(
        type=dataset_type,
        data_root=data_root,  # set by user
        pipeline=val_pipeline),
    sampler=dict(type='DefaultSampler', shuffle=False),
    persistent_workers=True)

test_dataloader = dict(
    batch_size=4,
    num_workers=8,
    dataset=dict(
        type=dataset_type,
        data_root=data_root,  # set by user
        pipeline=val_pipeline),
    sampler=dict(type='DefaultSampler', shuffle=False),
    persistent_workers=True)

# VIS_HOOK
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=5000,
        fixed_input=True,
        vis_kwargs_list=dict(type='GAN', name='fake_img'))
]

# METRICS
metrics = [
    dict(
        type='FrechetInceptionDistance',
        prefix='FID-50k',
        fake_nums=50000,
        real_nums=50000,
        inception_style='StyleGAN',
        sample_model='ema'),
    dict(type='PrecisionAndRecall', fake_nums=50000, prefix='PR-50K'),
    dict(type='PerceptualPathLength', fake_nums=50000, prefix='ppl-w')
]
# NOTE: config for save multi best checkpoints
# default_hooks = dict(
#     checkpoint=dict(
#         save_best=['FID-Full-50k/fid', 'IS-50k/is'],
#         rule=['less', 'greater']))
default_hooks = dict(checkpoint=dict(save_best='FID-50k/fid'))

val_evaluator = dict(metrics=metrics)
test_evaluator = dict(metrics=metrics)



# Copyright (c) OpenMMLab. All rights reserved.
from .mask_conv_module import MaskConvModule
from .partial_conv import PartialConv2d
from .pconv_decoder import PConvDecoder
from .pconv_encoder import PConvEncoder
from .pconv_encoder_decoder import PConvEncoderDecoder
from .pconv_inpaintor import PConvInpaintor

__all__ = [
    'PConvEncoder', 'PConvDecoder', 'PConvEncoderDecoder', 'PConvInpaintor',
    'MaskConvModule', 'PartialConv2d', 'MaskConvModule'
]



unet = dict(
    type='DenoisingUnet',
    image_size=256,
    in_channels=3,
    base_channels=256,
    resblocks_per_downsample=2,
    attention_res=(32, 16, 8),
    norm_cfg=dict(type='GN32', num_groups=32),
    dropout=0.0,
    num_classes=0,
    use_fp16=True,
    resblock_updown=True,
    attention_cfg=dict(
        type='MultiHeadAttentionBlock',
        num_heads=4,
        num_head_channels=64,
        use_new_attention_order=False),
    use_scale_shift_norm=True)

unet_ckpt_path = 'https://download.openmmlab.com/mmediting/synthesizers/disco/adm-u-cvt-rgb_finetuned_imagenet-256x256-f0f6e11b.pth'  # noqa
secondary_model_ckpt_path = 'https://download.openmmlab.com/mmediting/synthesizers/disco/secondary_model_imagenet_2.pth'  # noqa
pretrained_cfgs = dict(
    unet=dict(ckpt_path=unet_ckpt_path, prefix='unet'),
    secondary_model=dict(ckpt_path=secondary_model_ckpt_path, prefix=''))

secondary_model = dict(type='SecondaryDiffusionImageNet2')

diffusion_scheduler = dict(
    type='EditDDIMScheduler',
    variance_type='learned_range',
    beta_schedule='linear',
    clip_sample=False)

clip_models = [
    dict(type='ClipWrapper', clip_type='clip', name='ViT-B/32', jit=False),
    dict(type='ClipWrapper', clip_type='clip', name='ViT-B/16', jit=False),
    dict(type='ClipWrapper', clip_type='clip', name='RN50', jit=False)
]

model = dict(
    type='DiscoDiffusion',
    unet=unet,
    diffusion_scheduler=diffusion_scheduler,
    secondary_model=secondary_model,
    clip_models=clip_models,
    use_fp16=True,
    pretrained_cfgs=pretrained_cfgs)



# Copyright (c) OpenMMLab. All rights reserved.
from .contextual_attention import ContextualAttentionModule
from .contextual_attention_neck import ContextualAttentionNeck
from .deepfill_decoder import DeepFillDecoder
from .deepfill_disc import DeepFillv1Discriminators
from .deepfill_encoder import DeepFillEncoder
from .deepfill_refiner import DeepFillRefiner
from .deepfillv1 import DeepFillv1Inpaintor

__all__ = [
    'DeepFillEncoder', 'DeepFillDecoder', 'ContextualAttentionNeck',
    'DeepFillv1Inpaintor', 'ContextualAttentionModule',
    'DeepFillv1Discriminators', 'DeepFillRefiner'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .concat_visualizer import ConcatImageVisualizer
from .vis_backend import (PaviVisBackend, TensorboardVisBackend, VisBackend,
                          WandbVisBackend)
from .visualizer import Visualizer

__all__ = [
    'ConcatImageVisualizer', 'Visualizer', 'VisBackend', 'PaviVisBackend',
    'TensorboardVisBackend', 'WandbVisBackend'
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn.functional as F
from mmengine.model import BaseModule

from mmagic.registry import MODELS


@MODELS.register_module()
class DeepFillRefiner(BaseModule):
    """Refiner used in DeepFill model.

    This implementation follows:
    Generative Image Inpainting with Contextual Attention.

    Args:
        encoder_attention (dict): Config dict for encoder used in branch
            with contextual attention module.
        encoder_conv (dict): Config dict for encoder used in branch with
            just convolutional operation.
        dilation_neck (dict): Config dict for dilation neck in branch with
            just convolutional operation.
        contextual_attention (dict): Config dict for contextual attention
            neck.
        decoder (dict): Config dict for decoder used to fuse and decode
            features.
    """

    def __init__(self,
                 encoder_attention=dict(
                     type='DeepFillEncoder', encoder_type='stage2_attention'),
                 encoder_conv=dict(
                     type='DeepFillEncoder', encoder_type='stage2_conv'),
                 dilation_neck=dict(
                     type='GLDilationNeck',
                     in_channels=128,
                     act_cfg=dict(type='ELU')),
                 contextual_attention=dict(
                     type='ContextualAttentionNeck', in_channels=128),
                 decoder=dict(type='DeepFillDecoder', in_channels=256)):
        super().__init__()
        self.encoder_attention = MODELS.build(encoder_attention)
        self.encoder_conv = MODELS.build(encoder_conv)
        self.contextual_attention_neck = MODELS.build(contextual_attention)
        self.dilation_neck = MODELS.build(dilation_neck)
        self.decoder = MODELS.build(decoder)

    def forward(self, x, mask):
        """Forward Function.

        Args:
            x (torch.Tensor): Input tensor with shape of (n, c, h, w).
            mask (torch.Tensor): Input tensor with shape of (n, 1, h, w).

        Returns:
            torch.Tensor: Output tensor with shape of (n, c, h', w').
        """
        # conv branch
        encoder_dict = self.encoder_conv(x)
        conv_x = self.dilation_neck(encoder_dict['out'])

        # contextual attention branch
        attention_x = self.encoder_attention(x)['out']
        h_x, w_x = attention_x.shape[-2:]
        # resale mask to a smaller size
        resized_mask = F.interpolate(mask, size=(h_x, w_x))
        attention_x, offset = self.contextual_attention_neck(
            attention_x, resized_mask)

        # concat two branches
        x = torch.cat([conv_x, attention_x], dim=1)
        x = self.decoder(dict(out=x))

        return x, offset



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule
from mmengine.model.weight_init import kaiming_init, normal_init

from mmagic.models.archs import DepthwiseSeparableConvModule
from mmagic.registry import MODELS


class IndexedUpsample(BaseModule):
    """Indexed upsample module.

    Args:
        in_channels (int): Input channels.
        out_channels (int): Output channels.
        kernel_size (int, optional): Kernel size of the convolution layer.
            Defaults to 5.
        norm_cfg (dict, optional): Config dict for normalization layer.
            Defaults to dict(type='BN').
        conv_module (ConvModule | DepthwiseSeparableConvModule, optional):
            Conv module. Defaults to ConvModule.
        init_cfg (dict, optional): Initialization config dict. Default: None.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size=5,
                 norm_cfg=dict(type='BN'),
                 conv_module=ConvModule,
                 init_cfg: Optional[dict] = None):
        super().__init__(init_cfg=init_cfg)

        self.conv = conv_module(
            in_channels,
            out_channels,
            kernel_size,
            padding=(kernel_size - 1) // 2,
            norm_cfg=norm_cfg,
            act_cfg=dict(type='ReLU6'))

    def init_weights(self):
        """Init weights for the module."""
        if self.init_cfg is not None:
            super().init_weights()
        else:
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    kaiming_init(m, mode='fan_in', nonlinearity='leaky_relu')

    def forward(self, x, shortcut, dec_idx_feat=None):
        """Forward function.

        Args:
            x (Tensor): Input feature map with shape (N, C, H, W).
            shortcut (Tensor): The shortcut connection with shape
                (N, C, H', W').
            dec_idx_feat (Tensor, optional): The decode index feature map with
                shape (N, C, H', W'). Defaults to None.

        Returns:
            Tensor: Output tensor with shape (N, C, H', W').
        """
        if dec_idx_feat is not None:
            assert shortcut.dim() == 4, (
                'shortcut must be tensor with 4 dimensions')
            x = dec_idx_feat * F.interpolate(x, size=shortcut.shape[2:])
        out = torch.cat((x, shortcut), dim=1)
        return self.conv(out)


@MODELS.register_module()
class IndexNetDecoder(BaseModule):
    """Decoder for IndexNet.

    Please refer to https://arxiv.org/abs/1908.00672.

    Args:
        in_channels (int): Input channels of the decoder.
        kernel_size (int, optional): Kernel size of the convolution layer.
            Defaults to 5.
        norm_cfg (None | dict, optional): Config dict for normalization
            layer. Defaults to dict(type='BN').
        separable_conv (bool): Whether to use separable conv. Default: False.
        init_cfg (dict, optional): Initialization config dict. Default: None.
    """

    def __init__(self,
                 in_channels,
                 kernel_size=5,
                 norm_cfg=dict(type='BN'),
                 separable_conv=False,
                 init_cfg: Optional[dict] = None):
        # TODO: remove in_channels argument
        super().__init__(init_cfg=init_cfg)

        if separable_conv:
            conv_module = DepthwiseSeparableConvModule
        else:
            conv_module = ConvModule

        blocks_in_channels = [
            in_channels * 2, 96 * 2, 64 * 2, 32 * 2, 24 * 2, 16 * 2, 32 * 2
        ]
        blocks_out_channels = [96, 64, 32, 24, 16, 32, 32]

        self.decoder_layers = nn.ModuleList()
        for in_channel, out_channel in zip(blocks_in_channels,
                                           blocks_out_channels):
            self.decoder_layers.append(
                IndexedUpsample(in_channel, out_channel, kernel_size, norm_cfg,
                                conv_module))

        self.pred = nn.Sequential(
            conv_module(
                32,
                1,
                kernel_size,
                padding=(kernel_size - 1) // 2,
                norm_cfg=norm_cfg,
                act_cfg=dict(type='ReLU6')),
            nn.Conv2d(
                1, 1, kernel_size, padding=(kernel_size - 1) // 2, bias=False))

    def init_weights(self):
        """Init weights for the module."""
        if self.init_cfg is not None:
            super().init_weights()
        else:
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    std = math.sqrt(2. /
                                    (m.out_channels * m.kernel_size[0]**2))
                    normal_init(m, mean=0, std=std)

    def forward(self, inputs):
        """Forward function.

        Args:
            inputs (dict): Output dict of IndexNetEncoder.

        Returns:
            Tensor: Predicted alpha matte of the current batch.
        """
        shortcuts = reversed(inputs['shortcuts'])
        dec_idx_feat_list = reversed(inputs['dec_idx_feat_list'])
        out = inputs['out']

        group = (self.decoder_layers, shortcuts, dec_idx_feat_list)
        for decode_layer, shortcut, dec_idx_feat in zip(*group):
            out = decode_layer(out, shortcut, dec_idx_feat)

        out = self.pred(out)

        return out



# Copyright (c) OpenMMLab. All rights reserved.
from .log_processor import LogProcessor
from .multi_loops import MultiTestLoop, MultiValLoop

__all__ = ['MultiTestLoop', 'MultiValLoop', 'LogProcessor']



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Tuple

import torch


def get_ray_limits_box(rays_o: torch.Tensor, rays_d: torch.Tensor,
                       box_side_length: float
                       ) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Author: Petr Kellnhofer
    Intersects rays with the [-1, 1] NDC volume.
    Returns min and max distance of entry.
    Returns -1 for no intersection.
    https://www.scratchapixel.com/lessons/3d-basic-rendering/minimal-ray-tracer-rendering-simple-shapes/ray-box-intersection  # noqa

    Args:
        rays_o (torch.Tensor): The origin of each ray.
        rays_d (torch.Tensor): The direction vector of each ray.
        box_side_length (float): The side length of axis aligned
            bounding box (AABB).

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Start and end point
            for each ray. Both shape like (bz, res, res, 1).
    """
    o_shape = rays_o.shape
    rays_o = rays_o.detach().reshape(-1, 3)
    rays_d = rays_d.detach().reshape(-1, 3)

    bb_min = [
        -1 * (box_side_length / 2), -1 * (box_side_length / 2),
        -1 * (box_side_length / 2)
    ]
    bb_max = [
        1 * (box_side_length / 2), 1 * (box_side_length / 2),
        1 * (box_side_length / 2)
    ]
    bounds = torch.tensor([bb_min, bb_max],
                          dtype=rays_o.dtype,
                          device=rays_o.device)
    is_valid = torch.ones(rays_o.shape[:-1], dtype=bool, device=rays_o.device)

    # Precompute inverse for stability.
    invdir = 1 / rays_d
    sign = (invdir < 0).long()

    # Intersect with YZ plane.
    tmin = (bounds.index_select(0, sign[..., 0])[..., 0] -
            rays_o[..., 0]) * invdir[..., 0]
    tmax = (bounds.index_select(0, 1 - sign[..., 0])[..., 0] -
            rays_o[..., 0]) * invdir[..., 0]

    # Intersect with XZ plane.
    tymin = (bounds.index_select(0, sign[..., 1])[..., 1] -
             rays_o[..., 1]) * invdir[..., 1]
    tymax = (bounds.index_select(0, 1 - sign[..., 1])[..., 1] -
             rays_o[..., 1]) * invdir[..., 1]

    # Resolve parallel rays.
    is_valid[torch.logical_or(tmin > tymax, tymin > tmax)] = False

    # Use the shortest intersection.
    tmin = torch.max(tmin, tymin)
    tmax = torch.min(tmax, tymax)

    # Intersect with XY plane.
    tzmin = (bounds.index_select(0, sign[..., 2])[..., 2] -
             rays_o[..., 2]) * invdir[..., 2]
    tzmax = (bounds.index_select(0, 1 - sign[..., 2])[..., 2] -
             rays_o[..., 2]) * invdir[..., 2]

    # Resolve parallel rays.
    is_valid[torch.logical_or(tmin > tzmax, tzmin > tmax)] = False

    # Use the shortest intersection.
    tmin = torch.max(tmin, tzmin)
    tmax = torch.min(tmax, tzmax)

    # Mark invalid.
    tmin[torch.logical_not(is_valid)] = -1
    tmax[torch.logical_not(is_valid)] = -2

    return tmin.reshape(*o_shape[:-1], 1), tmax.reshape(*o_shape[:-1], 1)


def inverse_transform_sampling(bins: torch.Tensor,
                               weights: torch.Tensor,
                               n_importance: int,
                               deterministic: bool = False,
                               eps: float = 1e-5) -> torch.Tensor:
    """Sample `N_importance` samples from `bins` with distribution defined by
    `weights`.

    Args:
        bins (int): (N_points, N_samples+1) where N_samples is the number
            of coarse samples per ray - 2.
        weights (torch.Tensor): Weights shape like (N_points, N_samples-1).
        n_importance (int): The number of samples to draw from the
            distribution.
        deterministic (bool): Whether use deterministic sampling method.
            Defaults to False.
        eps (float): a small number to prevent division by zero.
            Defaults to 1e-5.

    Outputs:
        torch.Tensor: the sampled samples.
    """
    N_rays, N_samples_ = weights.shape
    # prevent division by zero (don't do inplace op!)
    weights = weights + eps
    # (N_rays, N_samples_)
    pdf = weights / torch.sum(weights, -1, keepdim=True)
    # (N_rays, N_samples), cumulative distribution function
    cdf = torch.cumsum(pdf, -1)
    # (N_rays, N_samples_+1)
    cdf = torch.cat([torch.zeros_like(cdf[:, :1]), cdf], -1)
    # padded to 0~1 inclusive

    if deterministic:
        u = torch.linspace(0, 1, n_importance, device=bins.device)
        u = u.expand(N_rays, n_importance)
    else:
        u = torch.rand(N_rays, n_importance, device=bins.device)
    u = u.contiguous()

    inds = torch.searchsorted(cdf, u, right=True)
    below = torch.clamp_min(inds - 1, 0)
    above = torch.clamp_max(inds, N_samples_)

    inds_sampled = torch.stack([below, above],
                               -1).view(N_rays, 2 * n_importance)
    cdf_g = torch.gather(cdf, 1, inds_sampled).view(N_rays, n_importance, 2)
    bins_g = torch.gather(bins, 1, inds_sampled).view(N_rays, n_importance, 2)

    denom = cdf_g[..., 1] - cdf_g[..., 0]
    # denom equals 0 means a bin has weight 0, in which case it will not
    # be sampled anyway, therefore any value for it is fine (set to 1 here)
    denom[denom < eps] = 1
    t = (u - cdf_g[..., 0]) / denom
    samples = bins_g[..., 0] + t * (bins_g[..., 1] - bins_g[..., 0])

    return samples


def linspace_batch(start: torch.Tensor, stop: torch.Tensor,
                   num: int) -> torch.Tensor:
    """Creates a tensor of shape [num, *start.shape] whose values are evenly
    spaced from start to end, inclusive.

    Replicates but the multi-dimensional behaviour of numpy.linspace in
    PyTorch.

    Args:
        start (torch.Tensor): The start point of each ray. Shape like
            (bz, res, res, 1).
        stop (torch.Tensor): The end point of each ray. Shape like
            (bz, res, res, 1).
        num (int): The number of points to sample.

    Returns:
        torch.Tensor: The sampled points. Shape like (num, bz, res, res, 1)
    """
    # create a tensor of 'num' steps from 0 to 1
    steps = torch.arange(
        num, dtype=torch.float32, device=start.device) / (
            num - 1)

    # reshape the 'steps' tensor to [-1, *([1]*start.ndim)] to allow for
    # broadcastings
    # - using 'steps.reshape([-1, *([1]*start.ndim)])' would be nice here but
    # torchscript cannot statically infer the expected size of a list in this
    # context, hence the code below
    for i in range(start.ndim):
        steps = steps.unsqueeze(-1)

    # the output starts at 'start' and increments until 'stop' in
    # each dimension
    out = start[None] + steps * (stop - start)[None]

    return out



# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule

from mmagic.registry import MODELS


@MODELS.register_module()
class GLEncoder(BaseModule):
    """Encoder used in Global&Local model.

    This implementation follows:
    Globally and locally Consistent Image Completion

    Args:
        norm_cfg (dict): Config dict to build norm layer.
        act_cfg (dict): Config dict for activation layer, "relu" by default.
    """

    def __init__(self, norm_cfg=None, act_cfg=dict(type='ReLU')):
        super().__init__()

        channel_list = [64, 128, 128, 256, 256, 256]
        kernel_size_list = [5, 3, 3, 3, 3, 3]
        stride_list = [1, 2, 1, 2, 1, 1]
        in_channels = 4
        for i in range(6):
            ks = kernel_size_list[i]
            padding = (ks - 1) // 2
            self.add_module(
                f'enc{i + 1}',
                ConvModule(
                    in_channels,
                    channel_list[i],
                    kernel_size=ks,
                    stride=stride_list[i],
                    padding=padding,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg))
            in_channels = channel_list[i]

    def forward(self, x):
        """Forward Function.

        Args:
            x (torch.Tensor): Input tensor with shape of (n, c, h, w).

        Returns:
            torch.Tensor: Output tensor with shape of (n, c, h', w').
        """
        for i in range(6):
            x = getattr(self, f'enc{i + 1}')(x)
        return x



# Copyright (c) OpenMMLab. All rights reserved.
"""Evaluation metrics based on each sample."""

from typing import List, Optional, Sequence

import torch.nn as nn
from mmengine.evaluator import BaseMetric
from mmengine.model import is_model_wrapper
from torch.utils.data.dataloader import DataLoader

from mmagic.registry import METRICS
from .metrics_utils import average, obtain_data


@METRICS.register_module()
class BaseSampleWiseMetric(BaseMetric):
    """Base sample wise metric of edit.

    Subclass must provide process function.

    Args:

        gt_key (str): Key of ground-truth. Default: 'gt_img'
        pred_key (str): Key of prediction. Default: 'pred_img'
        mask_key (str, optional): Key of mask, if mask_key is None, calculate
            all regions. Default: None
        collect_device (str): Device name used for collecting results from
            different ranks during distributed training. Must be 'cpu' or
            'gpu'. Defaults to 'cpu'.
        device (str): Device used to place torch tensors to compute metrics.
            Defaults to 'cpu'.
        prefix (str, optional): The prefix that will be added in the metric
            names to disambiguate homonymous metrics of different evaluators.
            If prefix is not provided in the argument, self.default_prefix
            will be used instead. Default: None
        scaling (float, optional): Scaling factor for final metric.
            E.g. scaling=100 means the final metric will be amplified by 100
            for output. Default: 1
    """

    SAMPLER_MODE = 'normal'
    sample_model = 'orig'  # TODO: low-level models only support origin model
    metric = None  # the name of metric

    def __init__(self,
                 gt_key: str = 'gt_img',
                 pred_key: str = 'pred_img',
                 mask_key: Optional[str] = None,
                 scaling=1,
                 device='cpu',
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None) -> None:
        assert self.metric is not None, (
            '\'metric\' must be defined for \'BaseSampleWiseMetric\'.')
        super().__init__(collect_device, prefix)

        self.gt_key = gt_key
        self.pred_key = pred_key
        self.mask_key = mask_key
        self.scaling = scaling
        self.device = device

        self.channel_order = 'BGR'

    def compute_metrics(self, results: List):
        """Compute the metrics from processed results.

        Args:
            results (List): The processed results of each batch.

        Returns:
            Dict: The computed metrics. The keys are the names of the metrics,
            and the values are corresponding results.
        """

        result = average(results, self.metric) * self.scaling

        return {self.metric: result}

    def process(self, data_batch: Sequence[dict],
                data_samples: Sequence[dict]) -> None:
        """Process one batch of data and predictions.

        Args:
            data_batch (Sequence[dict]): A batch of data
                from the dataloader.
            predictions (Sequence[dict]): A batch of outputs from
                the model.
        """

        for data in data_samples:
            prediction = data['output']

            gt = obtain_data(data, self.gt_key, self.device)
            pred = obtain_data(prediction, self.pred_key, self.device)
            if self.mask_key is not None:
                mask = obtain_data(data, self.mask_key)
                mask[mask != 0] = 1
            else:
                mask = 1 - pred * 0

            if len(gt.shape) <= 3:
                result = self.process_image(gt, pred, mask)
            else:
                result_sum = 0
                for i in range(gt.shape[0]):
                    result_sum += self.process_image(gt[i], pred[i], mask[i])
                result = result_sum / gt.shape[0]

            self.results.append({self.metric: result})

    def process_image(self, gt, pred, mask):
        raise NotImplementedError

    def evaluate(self) -> dict:
        assert hasattr(self, 'size'), (
            'Cannot find \'size\', please make sure \'self.prepare\' is '
            'called correctly.')
        return super().evaluate(self.size)

    def prepare(self, module: nn.Module, dataloader: DataLoader):
        self.size = len(dataloader.dataset)
        if is_model_wrapper(module):
            module = module.module
        self.data_preprocessor = module.data_preprocessor

    def get_metric_sampler(self, model: nn.Module, dataloader: DataLoader,
                           metrics) -> DataLoader:
        """Get sampler for normal metrics. Directly returns the dataloader.

        Args:
            model (nn.Module): Model to evaluate.
            dataloader (DataLoader): Dataloader for real images.
            metrics (List['GenMetric']): Metrics with the same sample mode.

        Returns:
            DataLoader: Default sampler for normal metrics.
        """
        return dataloader



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy

import numpy as np
import torch
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule

from mmagic.models.utils import get_module_device
from mmagic.registry import MODELS
from .wgan_gp_module import WGANNoiseTo2DFeat


@MODELS.register_module()
class WGANGPGenerator(BaseModule):
    r"""Generator for WGANGP.

    Implementation Details for WGANGP generator the same as training
    configuration (a) described in PGGAN paper:
    PROGRESSIVE GROWING OF GANS FOR IMPROVED QUALITY, STABILITY, AND VARIATION
    https://research.nvidia.com/sites/default/files/pubs/2017-10_Progressive-Growing-of/karras2018iclr-paper.pdf # noqa

    #. Adopt convolution architecture specified in appendix A.2;
    #. Use batchnorm in the generator except for the final output layer;
    #. Use ReLU in the generator except for the final output layer;
    #. Use Tanh in the last layer;
    #. Initialize all weights using He’s initializer.

    Args:
        noise_size (int): Size of the input noise vector.
        out_scale (int): Output scale for the generated image.
        conv_module_cfg (dict, optional): Config for the convolution
            module used in this generator. Defaults to None.
        upsample_cfg (dict, optional): Config for the upsampling operation.
            Defaults to None.
        init_cfg (dict, optional): Initialization config dict.
    """
    _default_channels_per_scale = {
        '4': 512,
        '8': 512,
        '16': 256,
        '32': 128,
        '64': 64,
        '128': 32
    }
    _default_conv_module_cfg = dict(
        conv_cfg=None,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=True,
        act_cfg=dict(type='ReLU'),
        norm_cfg=dict(type='BN'),
        order=('conv', 'norm', 'act'))

    _default_upsample_cfg = dict(type='nearest', scale_factor=2)

    def __init__(self,
                 noise_size,
                 out_scale,
                 conv_module_cfg=None,
                 upsample_cfg=None,
                 init_cfg=None):
        super().__init__(init_cfg=init_cfg)
        # set initial params
        self.noise_size = noise_size
        self.out_scale = out_scale
        self.conv_module_cfg = deepcopy(self._default_conv_module_cfg)
        if conv_module_cfg is not None:
            self.conv_module_cfg.update(conv_module_cfg)
        self.upsample_cfg = upsample_cfg if upsample_cfg else deepcopy(
            self._default_upsample_cfg)
        # set noise2feat head
        self.noise2feat = WGANNoiseTo2DFeat(
            self.noise_size, self._default_channels_per_scale['4'])
        # set conv_blocks
        self.conv_blocks = nn.ModuleList()
        self.conv_blocks.append(ConvModule(512, 512, **self.conv_module_cfg))

        log2scale = int(np.log2(self.out_scale))
        for i in range(3, log2scale + 1):
            self.conv_blocks.append(MODELS.build(self._default_upsample_cfg))
            self.conv_blocks.append(
                ConvModule(self._default_channels_per_scale[str(2**(i - 1))],
                           self._default_channels_per_scale[str(2**i)],
                           **self.conv_module_cfg))
            self.conv_blocks.append(
                ConvModule(self._default_channels_per_scale[str(2**i)],
                           self._default_channels_per_scale[str(2**i)],
                           **self.conv_module_cfg))
        self.to_rgb = ConvModule(
            self._default_channels_per_scale[str(self.out_scale)],
            kernel_size=1,
            out_channels=3,
            act_cfg=dict(type='Tanh'))

    def forward(self, noise, num_batches=0, return_noise=False):
        """Forward function.

        Args:
            noise (torch.Tensor | callable | None): You can directly give a
                batch of noise through a ``torch.Tensor`` or offer a callable
                function to sample a batch of noise data. Otherwise, the
                ``None`` indicates to use the default noise sampler.
            num_batches (int, optional): The number of batch size. Defaults to
                0.
            return_noise (bool, optional):  If True, ``noise_batch`` will be
                returned in a dict with ``fake_img``. Defaults to False.

        Returns:
            torch.Tensor | dict: If not ``return_noise``, only the output image
                will be returned. Otherwise, a dict contains ``fake_img`` and
                ``noise_batch`` will be returned.
        """
        # receive noise and conduct sanity check.
        if isinstance(noise, torch.Tensor):
            assert noise.shape[1] == self.noise_size
            assert noise.ndim == 2, ('The noise should be in shape of (n, c), '
                                     f'but got {noise.shape}')
            noise_batch = noise
        # receive a noise generator and sample noise.
        elif callable(noise):
            noise_generator = noise
            assert num_batches > 0
            noise_batch = noise_generator((num_batches, self.noise_size))
        # otherwise, we will adopt default noise sampler.
        else:
            assert num_batches > 0
            noise_batch = torch.randn((num_batches, self.noise_size))

        # dirty code for putting data on the right device
        noise_batch = noise_batch.to(get_module_device(self))

        # noise vector to 2D feature
        x = self.noise2feat(noise_batch)
        for conv in self.conv_blocks:
            x = conv(x)
        out_img = self.to_rgb(x)

        if return_noise:
            output = dict(fake_img=out_img, noise_batch=noise_batch)
            return output

        return out_img



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch
import torch.nn as nn
from mmcv.cnn import ConvModule


class ResidualBlockWithDropout(nn.Module):
    """Define a Residual Block with dropout layers.

    Ref:
    Deep Residual Learning for Image Recognition

    A residual block is a conv block with skip connections. A dropout layer is
    added between two common conv modules.

    Args:
        channels (int): Number of channels in the conv layer.
        padding_mode (str): The name of padding layer:
            'reflect' | 'replicate' | 'zeros'.
        norm_cfg (dict): Config dict to build norm layer. Default:
            `dict(type='IN')`.
        use_dropout (bool): Whether to use dropout layers. Default: True.
    """

    def __init__(self,
                 channels,
                 padding_mode,
                 norm_cfg=dict(type='BN'),
                 use_dropout=True):
        super().__init__()
        assert isinstance(norm_cfg, dict), ("'norm_cfg' should be dict, but"
                                            f'got {type(norm_cfg)}')
        assert 'type' in norm_cfg, "'norm_cfg' must have key 'type'"
        # We use norm layers in the residual block with dropout layers.
        # Only for IN, use bias to follow cyclegan's original implementation.
        use_bias = norm_cfg['type'] == 'IN'

        block = [
            ConvModule(
                in_channels=channels,
                out_channels=channels,
                kernel_size=3,
                padding=1,
                bias=use_bias,
                norm_cfg=norm_cfg,
                padding_mode=padding_mode)
        ]

        if use_dropout:
            block += [nn.Dropout(0.5)]

        block += [
            ConvModule(
                in_channels=channels,
                out_channels=channels,
                kernel_size=3,
                padding=1,
                bias=use_bias,
                norm_cfg=norm_cfg,
                act_cfg=None,
                padding_mode=padding_mode)
        ]

        self.block = nn.Sequential(*block)

    def forward(self, x):
        """Forward function. Add skip connections without final ReLU.

        Args:
            x (Tensor): Input tensor with shape (n, c, h, w).

        Returns:
            Tensor: Forward results.
        """
        out = x + self.block(x)
        return out


class GANImageBuffer:
    """This class implements an image buffer that stores previously generated
    images.

    This buffer allows us to update the discriminator using a history of
    generated images rather than the ones produced by the latest generator
    to reduce model oscillation.

    Args:
        buffer_size (int): The size of image buffer. If buffer_size = 0,
            no buffer will be created.
        buffer_ratio (float): The chance / possibility  to use the images
            previously stored in the buffer. Default: 0.5.
    """

    def __init__(self, buffer_size, buffer_ratio=0.5):
        self.buffer_size = buffer_size
        # create an empty buffer
        if self.buffer_size > 0:
            self.img_num = 0
            self.image_buffer = []
        self.buffer_ratio = buffer_ratio

    def query(self, images):
        """Query current image batch using a history of generated images.

        Args:
            images (Tensor): Current image batch without history information.
        """
        if self.buffer_size == 0:  # if the buffer size is 0, do nothing
            return images
        return_images = []
        for image in images:
            image = torch.unsqueeze(image.data, 0)
            # if the buffer is not full, keep inserting current images
            if self.img_num < self.buffer_size:
                self.img_num = self.img_num + 1
                self.image_buffer.append(image)
                return_images.append(image)
            else:
                use_buffer = np.random.random() < self.buffer_ratio
                # by self.buffer_ratio, the buffer will return a previously
                # stored image, and insert the current image into the buffer
                if use_buffer:
                    random_id = np.random.randint(0, self.buffer_size)
                    image_tmp = self.image_buffer[random_id].clone()
                    self.image_buffer[random_id] = image
                    return_images.append(image_tmp)
                # by (1 - self.buffer_ratio), the buffer will return the
                # current image
                else:
                    return_images.append(image)
        # collect all the images and return
        return_images = torch.cat(return_images, 0)
        return return_images



# Copyright (c) OpenMMLab. All rights reserved.
from .average_model import ExponentialMovingAverage, RampUpEMA
from .base_conditional_gan import BaseConditionalGAN
from .base_edit_model import BaseEditModel
from .base_gan import BaseGAN
from .base_mattor import BaseMattor
from .base_translation_model import BaseTranslationModel
from .basic_interpolator import BasicInterpolator
from .one_stage import OneStageInpaintor
from .two_stage import TwoStageInpaintor

__all__ = [
    'BaseEditModel',
    'BaseGAN',
    'BaseConditionalGAN',
    'BaseMattor',
    'BasicInterpolator',
    'BaseTranslationModel',
    'OneStageInpaintor',
    'TwoStageInpaintor',
    'ExponentialMovingAverage',
    'RampUpEMA',
]



# Copyright (c) OpenMMLab. All rights reserved.
import mmengine
import torch
from torch import nn

from mmagic.registry import MODELS
from .model_irse import Backbone


@MODELS.register_module('ArcFace')
class IDLossModel(nn.Module):
    """Face id loss model.

    Args:
        ir_se50_weights (str, optional): Url of ir-se50 weights.
            Defaults to None.
    """
    # ir se50 weight download link
    _ir_se50_url = 'https://gg0ltg.by.files.1drv.com/y4m3fNNszG03z9n8JQ7EhdtQKW8tQVQMFBisPVRgoXi_UfP8pKSSqv8RJNmHy2JampcPmEazo_Mx6NTFSqBpZmhPniROm9uNoghnzaavvYpxkCfiNmDH9YyIF3g-0nwt6bsjk2X80JDdL5z88OAblSDmB-kuQkWSWvA9BM3Xt8DHMCY8lO4HOQCZ5YWUtFyPAVwEyzTGDM-JRA5EJoN2bF1cg'  # noqa

    def __init__(self, ir_se50_weights=None):
        super(IDLossModel, self).__init__()
        mmengine.print_log('Loading ResNet ArcFace', 'current')
        self.facenet = Backbone(
            input_size=112, num_layers=50, drop_ratio=0.6, mode='ir_se')
        if ir_se50_weights is None:
            ir_se50_weights = self._ir_se50_url
        self.facenet.load_state_dict(
            torch.hub.load_state_dict_from_url(
                ir_se50_weights, map_location='cpu'))
        self.pool = torch.nn.AdaptiveAvgPool2d((256, 256))
        self.face_pool = torch.nn.AdaptiveAvgPool2d((112, 112))
        self.facenet = self.facenet.eval()

    def extract_feats(self, x):
        """Extracting face features.

        Args:
            x (torch.Tensor): Image tensor of faces.

        Returns:
            torch.Tensor: Face features.
        """
        if x.shape[2] != 256:
            x = self.pool(x)
        x = x[:, :, 35:223, 32:220]  # Crop interesting region
        x = self.face_pool(x)
        x_feats = self.facenet(x)
        return x_feats

    def forward(self, pred=None, gt=None):
        """Calculate face loss.

        Args:
            pred (torch.Tensor, optional): Predictions of face images.
                Defaults to None.
            gt (torch.Tensor, optional): Ground truth of face images.
                Defaults to None.

        Returns:
            Tuple(float, float): A tuple contain face similarity loss and
                improvement.
        """
        n_samples = gt.shape[0]
        y_feats = self.extract_feats(
            gt)  # Otherwise use the feature from there
        y_hat_feats = self.extract_feats(pred)
        y_feats = y_feats.detach()
        loss = 0
        sim_improvement = 0
        count = 0
        for i in range(n_samples):
            diff_target = y_hat_feats[i].dot(y_feats[i])
            loss += 1 - diff_target
            count += 1

        return loss / count, sim_improvement / count



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule
from mmengine.model.weight_init import constant_init

from mmagic.registry import MODELS
from .gca_module import GCAModule
from .resgca_enc import BasicBlock


class BasicBlockDec(BasicBlock):
    """Basic residual block for decoder.

    For decoder, we use ConvTranspose2d with kernel_size 4 and padding 1 for
    conv1. And the output channel of conv1 is modified from `out_channels` to
    `in_channels`.
    """

    def build_conv1(self, in_channels, out_channels, kernel_size, stride,
                    conv_cfg, norm_cfg, act_cfg, with_spectral_norm):
        """Build conv1 of the block.

        Args:
            in_channels (int): The input channels of the ConvModule.
            out_channels (int): The output channels of the ConvModule.
            kernel_size (int): The kernel size of the ConvModule.
            stride (int): The stride of the ConvModule. If stride is set to 2,
                then ``conv_cfg`` will be overwritten as
                ``dict(type='Deconv')`` and ``kernel_size`` will be overwritten
                as 4.
            conv_cfg (dict): The conv config of the ConvModule.
            norm_cfg (dict): The norm config of the ConvModule.
            act_cfg (dict): The activation config of the ConvModule.
            with_spectral_norm (bool): Whether use spectral norm.

        Returns:
            nn.Module: The built ConvModule.
        """
        if stride == 2:
            conv_cfg = dict(type='Deconv')
            kernel_size = 4
            padding = 1
        else:
            padding = kernel_size // 2

        return ConvModule(
            in_channels,
            in_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            with_spectral_norm=with_spectral_norm)

    def build_conv2(self, in_channels, out_channels, kernel_size, conv_cfg,
                    norm_cfg, with_spectral_norm):
        """Build conv2 of the block.

        Args:
            in_channels (int): The input channels of the ConvModule.
            out_channels (int): The output channels of the ConvModule.
            kernel_size (int): The kernel size of the ConvModule.
            conv_cfg (dict): The conv config of the ConvModule.
            norm_cfg (dict): The norm config of the ConvModule.
            with_spectral_norm (bool): Whether use spectral norm.

        Returns:
            nn.Module: The built ConvModule.
        """
        return ConvModule(
            in_channels,
            out_channels,
            kernel_size,
            stride=1,
            padding=kernel_size // 2,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=None,
            with_spectral_norm=with_spectral_norm)


@MODELS.register_module()
class ResNetDec(BaseModule):
    """ResNet decoder for image matting.

    This class is adopted from https://github.com/Yaoyi-Li/GCA-Matting.

    Args:
        block (str): Type of residual block. Currently only `BasicBlockDec` is
            implemented.
        layers (list[int]): Number of layers in each block.
        in_channels (int): Channel num of input features.
        kernel_size (int): Kernel size of the conv layers in the decoder.
        conv_cfg (dict): dictionary to construct convolution layer. If it is
            None, 2d convolution will be applied. Default: None.
        norm_cfg (dict): Config dict for normalization layer. "BN" by default.
        act_cfg (dict): Config dict for activation layer, "ReLU" by default.
        with_spectral_norm (bool): Whether use spectral norm after conv.
            Default: False.
        late_downsample (bool): Whether to adopt late downsample strategy,
            Default: False.
        init_cfg (dict, optional): Initialization config dict. Default: None.
    """

    def __init__(self,
                 block,
                 layers,
                 in_channels,
                 kernel_size=3,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(
                     type='LeakyReLU', negative_slope=0.2, inplace=True),
                 with_spectral_norm=False,
                 late_downsample=False,
                 init_cfg: Optional[dict] = None):
        super().__init__(init_cfg=init_cfg)
        if block == 'BasicBlockDec':
            block = BasicBlockDec
        else:
            raise NotImplementedError(f'{block} is not implemented.')

        self.kernel_size = kernel_size
        self.inplanes = in_channels
        self.midplanes = 64 if late_downsample else 32

        self.layer1 = self._make_layer(block, 256, layers[0], conv_cfg,
                                       norm_cfg, act_cfg, with_spectral_norm)
        self.layer2 = self._make_layer(block, 128, layers[1], conv_cfg,
                                       norm_cfg, act_cfg, with_spectral_norm)
        self.layer3 = self._make_layer(block, 64, layers[2], conv_cfg,
                                       norm_cfg, act_cfg, with_spectral_norm)
        self.layer4 = self._make_layer(block, self.midplanes, layers[3],
                                       conv_cfg, norm_cfg, act_cfg,
                                       with_spectral_norm)

        self.conv1 = ConvModule(
            self.midplanes,
            32,
            4,
            stride=2,
            padding=1,
            conv_cfg=dict(type='Deconv'),
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            with_spectral_norm=with_spectral_norm)

        self.conv2 = ConvModule(
            32,
            1,
            self.kernel_size,
            padding=self.kernel_size // 2,
            act_cfg=None)

    def init_weights(self):
        """Init weights for the module."""
        if self.init_cfg is not None:
            super().init_weights()
        else:
            for m in self.modules():
                if isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                    constant_init(m.weight, 1)
                    constant_init(m.bias, 0)

        # Zero-initialize the last BN in each residual branch, so that the
        # residual branch starts with zeros, and each residual block behaves
        # like an identity. This improves the model by 0.2~0.3% according to
        # https://arxiv.org/abs/1706.02677
        for m in self.modules():
            if isinstance(m, BasicBlockDec):
                constant_init(m.conv2.bn.weight, 0)

    def _make_layer(self, block, planes, num_blocks, conv_cfg, norm_cfg,
                    act_cfg, with_spectral_norm):
        upsample = nn.Sequential(
            nn.UpsamplingNearest2d(scale_factor=2),
            ConvModule(
                self.inplanes,
                planes * block.expansion,
                1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None,
                with_spectral_norm=with_spectral_norm))

        layers = [
            block(
                self.inplanes,
                planes,
                kernel_size=self.kernel_size,
                stride=2,
                interpolation=upsample,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
                with_spectral_norm=with_spectral_norm)
        ]
        self.inplanes = planes * block.expansion
        for _ in range(1, num_blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    kernel_size=self.kernel_size,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                    with_spectral_norm=with_spectral_norm))

        return nn.Sequential(*layers)

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (N, C, H, W).

        Returns:
            Tensor: Output tensor.
        """
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.conv1(x)
        x = self.conv2(x)

        return x


@MODELS.register_module()
class ResShortcutDec(ResNetDec):
    """ResNet decoder for image matting with shortcut connection.

    ::

        feat1 --------------------------- conv2 --- out
                                       |
        feat2 ---------------------- conv1
                                  |
        feat3 ----------------- layer4
                             |
        feat4 ------------ layer3
                        |
        feat5 ------- layer2
                   |
        out ---  layer1

    Args:
        block (str): Type of residual block. Currently only `BasicBlockDec` is
            implemented.
        layers (list[int]): Number of layers in each block.
        in_channels (int): Channel number of input features.
        kernel_size (int): Kernel size of the conv layers in the decoder.
        conv_cfg (dict): Dictionary to construct convolution layer. If it is
            None, 2d convolution will be applied. Default: None.
        norm_cfg (dict): Config dict for normalization layer. "BN" by default.
        act_cfg (dict): Config dict for activation layer, "ReLU" by default.
        late_downsample (bool): Whether to adopt late downsample strategy,
            Default: False.
    """

    def forward(self, inputs):
        """Forward function of resnet shortcut decoder.

        Args:
            inputs (dict): Output dictionary of the ResNetEnc containing:

              - out (Tensor): Output of the ResNetEnc.
              - feat1 (Tensor): Shortcut connection from input image.
              - feat2 (Tensor): Shortcut connection from conv2 of ResNetEnc.
              - feat3 (Tensor): Shortcut connection from layer1 of ResNetEnc.
              - feat4 (Tensor): Shortcut connection from layer2 of ResNetEnc.
              - feat5 (Tensor): Shortcut connection from layer3 of ResNetEnc.

        Returns:
            Tensor: Output tensor.
        """
        feat1 = inputs['feat1']
        feat2 = inputs['feat2']
        feat3 = inputs['feat3']
        feat4 = inputs['feat4']
        feat5 = inputs['feat5']
        x = inputs['out']

        x = self.layer1(x) + feat5
        x = self.layer2(x) + feat4
        x = self.layer3(x) + feat3
        x = self.layer4(x) + feat2
        x = self.conv1(x) + feat1
        x = self.conv2(x)

        return x


@MODELS.register_module()
class ResGCADecoder(ResShortcutDec):
    """ResNet decoder with shortcut connection and gca module.

    ::

        feat1 ---------------------------------------- conv2 --- out
                                                    |
        feat2 ----------------------------------- conv1
                                               |
        feat3 ------------------------------ layer4
                                          |
        feat4, img_feat -- gca_module - layer3
                        |
        feat5 ------- layer2
                   |
        out ---  layer1

    * gca module also requires unknown tensor generated by trimap which is \
    ignored in the above graph.

    Args:
        block (str): Type of residual block. Currently only `BasicBlockDec` is
            implemented.
        layers (list[int]): Number of layers in each block.
        in_channels (int): Channel number of input features.
        kernel_size (int): Kernel size of the conv layers in the decoder.
        conv_cfg (dict): Dictionary to construct convolution layer. If it is
            None, 2d convolution will be applied. Default: None.
        norm_cfg (dict): Config dict for normalization layer. "BN" by default.
        act_cfg (dict): Config dict for activation layer, "ReLU" by default.
        with_spectral_norm (bool): Whether use spectral norm. Default: False.
        late_downsample (bool): Whether to adopt late downsample strategy,
            Default: False.
    """

    def __init__(self,
                 block,
                 layers,
                 in_channels,
                 kernel_size=3,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(
                     type='LeakyReLU', negative_slope=0.2, inplace=True),
                 with_spectral_norm=False,
                 late_downsample=False):
        super().__init__(block, layers, in_channels, kernel_size, conv_cfg,
                         norm_cfg, act_cfg, with_spectral_norm,
                         late_downsample)
        self.gca = GCAModule(128, 128)

    def forward(self, inputs):
        """Forward function of resnet shortcut decoder.

        Args:
            inputs (dict): Output dictionary of the ResGCAEncoder containing:

              - out (Tensor): Output of the ResGCAEncoder.
              - feat1 (Tensor): Shortcut connection from input image.
              - feat2 (Tensor): Shortcut connection from conv2 of \
                    ResGCAEncoder.
              - feat3 (Tensor): Shortcut connection from layer1 of \
                    ResGCAEncoder.
              - feat4 (Tensor): Shortcut connection from layer2 of \
                    ResGCAEncoder.
              - feat5 (Tensor): Shortcut connection from layer3 of \
                    ResGCAEncoder.
              - img_feat (Tensor): Image feature extracted by guidance head.
              - unknown (Tensor): Unknown tensor generated by trimap.

        Returns:
            Tensor: Output tensor.
        """
        img_feat = inputs['img_feat']
        unknown = inputs['unknown']
        feat1 = inputs['feat1']
        feat2 = inputs['feat2']
        feat3 = inputs['feat3']
        feat4 = inputs['feat4']
        feat5 = inputs['feat5']
        x = inputs['out']

        x = self.layer1(x) + feat5
        x = self.layer2(x) + feat4
        x = self.gca(img_feat, x, unknown)
        x = self.layer3(x) + feat3
        x = self.layer4(x) + feat2
        x = self.conv1(x) + feat1
        x = self.conv2(x)

        return x



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import Callable, List, Union

from mmengine import FileClient
from mmengine.dataset import BaseDataset

from mmagic.registry import DATASETS


@DATASETS.register_module()
class DreamBoothDataset(BaseDataset):
    """Dataset for DreamBooth.

    Args:
        data_root (str): Path to the data root.
        concept_dir (str): Path to the concept images.
        prompt (str): Prompt of the concept.
        pipeline (list[dict | callable]): A sequence of data transforms.
    """

    def __init__(self,
                 data_root: str,
                 concept_dir: str,
                 prompt: str,
                 pipeline: List[Union[dict, Callable]] = []):

        data_prefix = dict(img_path=concept_dir)

        self.prompt = prompt

        super().__init__(
            data_root=data_root, data_prefix=data_prefix, pipeline=pipeline)

    def load_data_list(self) -> list:
        """Load data list from concept_dir and class_dir."""
        data_list = []

        img_dir = self.data_prefix['img_path']
        file_client = FileClient.infer_client(uri=img_dir)
        img_dir = osp.abspath(img_dir)

        for data_name in file_client.list_dir_or_file(img_dir, list_dir=False):
            data_info = dict(
                img_path=file_client.join_path(img_dir, data_name),
                prompt=self.prompt)
            data_list.append(data_info)

        return data_list



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy

import numpy as np
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule

from mmagic.registry import MODELS
from .wgan_gp_module import ConvLNModule, WGANDecisionHead


@MODELS.register_module()
class WGANGPDiscriminator(BaseModule):
    r"""Discriminator for WGANGP.

    Implementation Details for WGANGP discriminator the same as training
    configuration (a) described in PGGAN paper:
    PROGRESSIVE GROWING OF GANS FOR IMPROVED QUALITY, STABILITY, AND VARIATION
    https://research.nvidia.com/sites/default/files/pubs/2017-10_Progressive-Growing-of/karras2018iclr-paper.pdf # noqa

    #. Adopt convolution architecture specified in appendix A.2;
    #. Add layer normalization to all conv3x3 and conv4x4 layers;
    #. Use LeakyReLU in the discriminator except for the final output layer;
    #. Initialize all weights using He’s initializer.

    Args:
        in_channel (int): The channel number of the input image.
        in_scale (int): The scale of the input image.
        conv_module_cfg (dict, optional): Config for the convolution module
            used in this discriminator. Defaults to None.
        init_cfg (dict, optional): Initialization config dict.
    """
    _default_channels_per_scale = {
        '4': 512,
        '8': 512,
        '16': 256,
        '32': 128,
        '64': 64,
        '128': 32
    }
    _default_conv_module_cfg = dict(
        conv_cfg=None,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=True,
        act_cfg=dict(type='LeakyReLU', negative_slope=0.2),
        norm_cfg=dict(type='LN2d'),
        order=('conv', 'norm', 'act'))

    _default_upsample_cfg = dict(type='nearest', scale_factor=2)

    def __init__(self,
                 in_channel,
                 in_scale,
                 conv_module_cfg=None,
                 init_cfg=None):
        super().__init__(init_cfg=init_cfg)
        # set initial params
        self.in_channel = in_channel
        self.in_scale = in_scale
        self.conv_module_cfg = deepcopy(self._default_conv_module_cfg)
        if conv_module_cfg is not None:
            self.conv_module_cfg.update(conv_module_cfg)
        # set from_rgb head
        self.from_rgb = ConvModule(
            3,
            kernel_size=1,
            out_channels=self._default_channels_per_scale[str(self.in_scale)],
            act_cfg=dict(type='LeakyReLU', negative_slope=0.2))
        # set conv_blocks
        self.conv_blocks = nn.ModuleList()

        log2scale = int(np.log2(self.in_scale))
        for i in range(log2scale, 2, -1):
            self.conv_blocks.append(
                ConvLNModule(
                    self._default_channels_per_scale[str(2**i)],
                    self._default_channels_per_scale[str(2**i)],
                    feature_shape=(self._default_channels_per_scale[str(2**i)],
                                   2**i, 2**i),
                    **self.conv_module_cfg))
            self.conv_blocks.append(
                ConvLNModule(
                    self._default_channels_per_scale[str(2**i)],
                    self._default_channels_per_scale[str(2**(i - 1))],
                    feature_shape=(self._default_channels_per_scale[str(
                        2**(i - 1))], 2**i, 2**i),
                    **self.conv_module_cfg))
            self.conv_blocks.append(nn.AvgPool2d(kernel_size=2, stride=2))
        self.decision = WGANDecisionHead(
            self._default_channels_per_scale['4'],
            self._default_channels_per_scale['4'],
            1,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.2),
            norm_cfg=self.conv_module_cfg['norm_cfg'])

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor): Fake or real image tensor.

        Returns:
            torch.Tensor: Prediction for the reality of the input image.
        """
        # noise vector to 2D feature
        x = self.from_rgb(x)
        for conv in self.conv_blocks:
            x = conv(x)
        x = self.decision(x)
        return x



# Copyright (c) OpenMMLab. All rights reserved.
import random

import mmengine
import numpy as np
import torch
import torch.nn as nn
from mmengine.model import BaseModule
from mmengine.runner.amp import autocast
from mmengine.runner.checkpoint import _load_checkpoint_with_prefix

from mmagic.registry import MODELS
from ...utils import get_module_device
from ..pggan import PixelNorm
from ..stylegan1 import (ConstantInput, EqualLinearActModule, get_mean_latent,
                         style_mixing)
from .stylegan2_modules import ModulatedStyleConv, ModulatedToRGB


@MODELS.register_module('StyleGANv2Generator')
@MODELS.register_module()
class StyleGAN2Generator(BaseModule):
    r"""StyleGAN2 Generator.

    In StyleGAN2, we use a static architecture composing of a style mapping
    module and number of convolutional style blocks. More details can be found
    in: Analyzing and Improving the Image Quality of StyleGAN CVPR2020.

    You can load pretrained model through passing information into
    ``pretrained`` argument. We have already offered official weights as
    follows:

    - stylegan2-ffhq-config-f: https://download.openmmlab.com/mmediting/stylegan2/official_weights/stylegan2-ffhq-config-f-official_20210327_171224-bce9310c.pth  # noqa
    - stylegan2-horse-config-f: https://download.openmmlab.com/mmediting/stylegan2/official_weights/stylegan2-horse-config-f-official_20210327_173203-ef3e69ca.pth  # noqa
    - stylegan2-car-config-f: https://download.openmmlab.com/mmediting/stylegan2/official_weights/stylegan2-car-config-f-official_20210327_172340-8cfe053c.pth  # noqa
    - stylegan2-cat-config-f: https://download.openmmlab.com/mmediting/stylegan2/official_weights/stylegan2-cat-config-f-official_20210327_172444-15bc485b.pth  # noqa
    - stylegan2-church-config-f: https://download.openmmlab.com/mmediting/stylegan2/official_weights/stylegan2-church-config-f-official_20210327_172657-1d42b7d1.pth  # noqa

    If you want to load the ema model, you can just use following codes:

    .. code-block:: python

        # ckpt_http is one of the valid path from http source
        generator = StyleGANv2Generator(1024, 512,
                                        pretrained=dict(
                                            ckpt_path=ckpt_http,
                                            prefix='generator_ema'))

    Of course, you can also download the checkpoint in advance and set
    ``ckpt_path`` with local path. If you just want to load the original
    generator (not the ema model), please set the prefix with 'generator'.

    Note that our implementation allows to generate BGR image, while the
    original StyleGAN2 outputs RGB images by default. Thus, we provide
    ``bgr2rgb`` argument to convert the image space.

    Args:
        out_size (int): The output size of the StyleGAN2 generator.
        style_channels (int): The number of channels for style code.
        out_channels (int): The number of channels for output. Defaults to 3.
        noise_size (int, optional): The size of (number of channels) the input
            noise. If not passed, will be set the same value as
            :attr:`style_channels`. Defaults to None.
        cond_size (int, optional): The size of the conditional input. If not
            passed or less than 1, no conditional embedding will be used.
            Defaults to None.
        cond_mapping_channels (int, optional): The channels of the
            conditional mapping layers. If not passed, will use the same value
            as :attr:`style_channels`. Defaults to None.
        num_mlps (int, optional): The number of MLP layers. Defaults to 8.
        channel_multiplier (int, optional): The multiplier factor for the
            channel number. Defaults to 2.
        blur_kernel (list, optional): The blurry kernel. Defaults
            to [1, 3, 3, 1].
        lr_mlp (float, optional): The learning rate for the style mapping
            layer. Defaults to 0.01.
        default_style_mode (str, optional): The default mode of style mixing.
            In training, we adopt mixing style mode in default. However, in the
            evaluation, we use 'single' style mode. `['mix', 'single']` are
            currently supported. Defaults to 'mix'.
        eval_style_mode (str, optional): The evaluation mode of style mixing.
            Defaults to 'single'.
        mix_prob (float, optional): Mixing probability. The value should be
            in range of [0, 1]. Defaults to ``0.9``.
        update_mean_latent_with_ema (bool, optional): Whether update mean
            latent code (w) with EMA. Defaults to False.
        w_avg_beta (float, optional): The value used for update `w_avg`.
            Defaults to 0.998.
        num_fp16_scales (int, optional): The number of resolutions to use auto
            fp16 training. Different from ``fp16_enabled``, this argument
            allows users to adopt FP16 training only in several blocks.
            This behaviour is much more similar to the official implementation
            by Tero. Defaults to 0.
        fp16_enabled (bool, optional): Whether to use fp16 training in this
            module. If this flag is `True`, the whole module will be wrapped
            with ``auto_fp16``. Defaults to False.
        pretrained (dict | None, optional): Information for pretrained models.
            The necessary key is 'ckpt_path'. Besides, you can also provide
            'prefix' to load the generator part from the whole state dict.
            Defaults to None.
    """

    def __init__(self,
                 out_size,
                 style_channels,
                 out_channels=3,
                 noise_size=None,
                 cond_size=None,
                 cond_mapping_channels=None,
                 num_mlps=8,
                 channel_multiplier=2,
                 blur_kernel=[1, 3, 3, 1],
                 lr_mlp=0.01,
                 default_style_mode='mix',
                 eval_style_mode='single',
                 norm_eps=1e-6,
                 mix_prob=0.9,
                 update_mean_latent_with_ema=False,
                 w_avg_beta=0.998,
                 num_fp16_scales=0,
                 fp16_enabled=False,
                 bgr2rgb=False,
                 pretrained=None,
                 fixed_noise=False):
        super().__init__()
        self.out_size = out_size
        self.style_channels = style_channels
        self.out_channels = out_channels
        self.num_mlps = num_mlps
        self.channel_multiplier = channel_multiplier
        self.lr_mlp = lr_mlp
        self._default_style_mode = default_style_mode
        self.default_style_mode = default_style_mode
        self.eval_style_mode = eval_style_mode
        self.mix_prob = mix_prob
        self.num_fp16_scales = num_fp16_scales
        self.fp16_enabled = fp16_enabled
        self.bgr2rgb = bgr2rgb

        self.noise_size = style_channels if noise_size is None else noise_size

        self.cond_size = cond_size
        if self.cond_size is not None and self.cond_size > 0:
            cond_mapping_channels = style_channels \
                if cond_mapping_channels is None else cond_mapping_channels
            self.embed = EqualLinearActModule(cond_size, cond_mapping_channels)
            # NOTE: conditional input is passed, do 2nd moment norm for
            # embedding and noise input respectively, therefore mapping layer
            # start with FC layer
            mapping_layers = []
        else:
            cond_mapping_channels = 0
            # NOTE: conditional input is not passed, put 2nd moment norm at
            # the start of mapping layers
            mapping_layers = [PixelNorm(eps=norm_eps)]
        in_feat = cond_mapping_channels + self.noise_size

        # define pixel norm
        self.pixel_norm = PixelNorm(eps=norm_eps)

        # define style mapping layers
        for idx in range(num_mlps):
            mapping_layers.append(
                EqualLinearActModule(
                    in_feat if idx == 0 else style_channels,
                    style_channels,
                    equalized_lr_cfg=dict(lr_mul=lr_mlp, gain=1.),
                    act_cfg=dict(type='fused_bias')))

        self.style_mapping = nn.Sequential(*mapping_layers)

        self.channels = {
            4: 512,
            8: 512,
            16: 512,
            32: 512,
            64: 256 * channel_multiplier,
            128: 128 * channel_multiplier,
            256: 64 * channel_multiplier,
            512: 32 * channel_multiplier,
            1024: 16 * channel_multiplier,
        }

        # constant input layer
        self.constant_input = ConstantInput(self.channels[4])
        # 4x4 stage
        self.conv1 = ModulatedStyleConv(
            self.channels[4],
            self.channels[4],
            kernel_size=3,
            style_channels=style_channels,
            blur_kernel=blur_kernel,
            fp16_enabled=fp16_enabled,
            fixed_noise=fixed_noise)
        self.to_rgb1 = ModulatedToRGB(
            self.channels[4],
            style_channels,
            out_channels=out_channels,
            upsample=False,
            fp16_enabled=fp16_enabled)

        # generator backbone (8x8 --> higher resolutions)
        self.log_size = int(np.log2(self.out_size))

        self.convs = nn.ModuleList()
        self.upsamples = nn.ModuleList()
        self.to_rgbs = nn.ModuleList()

        blk_in_channels_ = self.channels[4]  # in channels of the conv blocks

        for i in range(3, self.log_size + 1):
            blk_out_channels_ = self.channels[2**i]

            # If `fp16_enabled` is True, all of layers will be run in auto
            # FP16. In the case of `num_fp16_scales` > 0, only partial
            # layers will be run in fp16.
            _use_fp16 = (self.log_size - i) < num_fp16_scales or fp16_enabled

            self.convs.append(
                ModulatedStyleConv(
                    blk_in_channels_,
                    blk_out_channels_,
                    3,
                    style_channels,
                    upsample=True,
                    blur_kernel=blur_kernel,
                    fp16_enabled=_use_fp16))
            self.convs.append(
                ModulatedStyleConv(
                    blk_out_channels_,
                    blk_out_channels_,
                    3,
                    style_channels,
                    upsample=False,
                    blur_kernel=blur_kernel,
                    fp16_enabled=_use_fp16))
            self.to_rgbs.append(
                ModulatedToRGB(
                    blk_out_channels_,
                    style_channels,
                    out_channels=out_channels,
                    upsample=True,
                    fp16_enabled=_use_fp16))  # set to global fp16

            blk_in_channels_ = blk_out_channels_

        self.num_latents = self.log_size * 2 - 2
        self.num_injected_noises = self.num_latents - 1

        # register buffer for injected noises
        for layer_idx in range(self.num_injected_noises):
            res = (layer_idx + 5) // 2
            shape = [1, 1, 2**res, 2**res]
            self.register_buffer(f'injected_noise_{layer_idx}',
                                 torch.randn(*shape))

        if (self.cond_size is not None
                and self.cond_size > 0) or update_mean_latent_with_ema:
            # Due to `get_mean_latent` cannot handle conditional input,
            # assign avg style code here and update with EMA.
            self.register_buffer('w_avg', torch.zeros([style_channels]))
            self.w_avg_beta = w_avg_beta
            mmengine.print_log('Mean latent code (w) is updated with EMA.')

        if pretrained is not None:
            self._load_pretrained_model(**pretrained)

    def _load_pretrained_model(self,
                               ckpt_path,
                               prefix='',
                               map_location='cpu',
                               strict=True):
        state_dict = _load_checkpoint_with_prefix(prefix, ckpt_path,
                                                  map_location)
        self.load_state_dict(state_dict, strict=strict)
        mmengine.print_log(f'Load pretrained model from {ckpt_path}')

    def train(self, mode=True):
        if mode:
            if self.default_style_mode != self._default_style_mode:
                mmengine.print_log(
                    f'Switch to train style mode: {self._default_style_mode}')
            self.default_style_mode = self._default_style_mode

        else:
            if self.default_style_mode != self.eval_style_mode:
                mmengine.print_log(
                    f'Switch to evaluation style mode: {self.eval_style_mode}')
            self.default_style_mode = self.eval_style_mode

        return super(StyleGAN2Generator, self).train(mode)

    def make_injected_noise(self):
        """make noises that will be injected into feature maps.

        Returns:
            list[Tensor]: List of layer-wise noise tensor.
        """
        device = get_module_device(self)

        noises = [torch.randn(1, 1, 2**2, 2**2, device=device)]

        for i in range(3, self.log_size + 1):
            for _ in range(2):
                noises.append(torch.randn(1, 1, 2**i, 2**i, device=device))

        return noises

    def get_mean_latent(self, num_samples=4096, **kwargs):
        """Get mean latent of W space in this generator.

        Args:
            num_samples (int, optional): Number of sample times. Defaults
                to 4096.

        Returns:
            Tensor: Mean latent of this generator.
        """
        if hasattr(self, 'w_avg'):
            mmengine.print_log('Get latent code (w) which is updated by EMA.')
            return self.w_avg
        return get_mean_latent(self, num_samples, **kwargs)

    def style_mixing(self,
                     n_source,
                     n_target,
                     inject_index=1,
                     truncation_latent=None,
                     truncation=0.7):
        return style_mixing(
            self,
            n_source=n_source,
            n_target=n_target,
            inject_index=inject_index,
            truncation=truncation,
            truncation_latent=truncation_latent,
            style_channels=self.style_channels)

    # @auto_fp16()
    def forward(self,
                styles,
                label=None,
                num_batches=-1,
                return_noise=False,
                return_latents=False,
                inject_index=None,
                truncation=1,
                truncation_latent=None,
                input_is_latent=False,
                injected_noise=None,
                add_noise=True,
                randomize_noise=True,
                update_ws=False,
                return_features=False,
                feat_idx=5,
                return_latent_only=False):
        """Forward function.

        This function has been integrated with the truncation trick. Please
        refer to the usage of `truncation` and `truncation_latent`.

        Args:
            styles (torch.Tensor | list[torch.Tensor] | callable | None): In
                StyleGAN2, you can provide noise tensor or latent tensor. Given
                a list containing more than one noise or latent tensors, style
                mixing trick will be used in training. Of course, You can
                directly give a batch of noise through a ``torch.Tensor`` or
                offer a callable function to sample a batch of noise data.
                Otherwise, the ``None`` indicates to use the default noise
                sampler.
            label (torch.Tensor, optional): Conditional inputs for the
                generator. Defaults to None.
            num_batches (int, optional): The number of batch size.
                Defaults to 0.
            return_noise (bool, optional): If True, ``noise_batch`` will be
                returned in a dict with ``fake_img``. Defaults to False.
            return_latents (bool, optional): If True, ``latent`` will be
                returned in a dict with ``fake_img``. Defaults to False.
            inject_index (int | None, optional): The index number for mixing
                style codes. Defaults to None.
            truncation (float, optional): Truncation factor. Give value less
                than 1., the truncation trick will be adopted. Defaults to 1.
            truncation_latent (torch.Tensor, optional): Mean truncation latent.
                Defaults to None.
            input_is_latent (bool, optional): If `True`, the input tensor is
                the latent tensor. Defaults to False.
            injected_noise (torch.Tensor | None, optional): Given a tensor, the
                random noise will be fixed as this input injected noise.
                Defaults to None.
            add_noise (bool): Whether apply noise injection. Defaults to True.
            randomize_noise (bool, optional): If `False`, images are sampled
                with the buffered noise tensor injected to the style conv
                block. Defaults to True.
            update_ws (bool): Whether update latent code with EMA. Only work
                when `w_avg` is registered. Defaults to False.

        Returns:
            torch.Tensor | dict: Generated image tensor or dictionary \
                containing more data.
        """
        # device = styles.device
        input_dim = self.style_channels if input_is_latent else self.noise_size
        # receive noise and conduct sanity check.
        if isinstance(styles, torch.Tensor):
            assert styles.shape[1] == input_dim
            styles = [styles]
        elif mmengine.is_seq_of(styles, torch.Tensor):
            for t in styles:
                assert t.shape[-1] == input_dim
        # receive a noise generator and sample noise.
        elif callable(styles):
            device = get_module_device(self)
            noise_generator = styles
            assert num_batches > 0
            if self.default_style_mode == 'mix' and random.random(
            ) < self.mix_prob:
                styles = [
                    noise_generator((num_batches, input_dim)) for _ in range(2)
                ]
            else:
                styles = [noise_generator((num_batches, input_dim))]
            styles = [s.to(device) for s in styles]
        # otherwise, we will adopt default noise sampler.
        else:
            device = get_module_device(self)
            assert num_batches > 0 and not input_is_latent
            if self.default_style_mode == 'mix' and random.random(
            ) < self.mix_prob:
                styles = [
                    torch.randn((num_batches, input_dim)) for _ in range(2)
                ]
            else:
                styles = [torch.randn((num_batches, input_dim))]
            styles = [s.to(device) for s in styles]

        # no amp for style-mapping and condition-embedding
        if not input_is_latent:
            noise_batch = styles
            if self.cond_size is not None and self.cond_size > 0:
                assert label is not None, (
                    '\'cond_channels\' is not None, \'cond\' must be passed.')
                assert label.shape[1] == self.cond_size
                embedding = self.embed(label)
                # NOTE: If conditional input is passed, do norm for cond
                # embedding and noise input respectively
                # do pixel_norm (2nd_moment_norm) to cond embedding
                embedding = self.pixel_norm(embedding)
                # do pixel_norm (2nd_moment_norm) to noise input
                styles = [self.pixel_norm(s) for s in styles]

            styles_list = []
            for s in styles:
                if self.cond_size is not None and self.cond_size > 0:
                    s = torch.cat([s, embedding], dim=1)
                styles_list.append(self.style_mapping(s))

            styles = styles_list
        else:
            noise_batch = None

        # update w_avg during training, if need
        if hasattr(self, 'w_avg') and self.training and update_ws:
            # only update w_avg with the first style code
            self.w_avg.copy_(styles[0].detach().mean(
                dim=0).lerp(self.w_avg, self.w_avg_beta))

        if injected_noise is None:
            if randomize_noise:
                injected_noise = [None] * self.num_injected_noises
            else:
                injected_noise = [
                    getattr(self, f'injected_noise_{i}')
                    for i in range(self.num_injected_noises)
                ]
        # use truncation trick
        if truncation < 1:
            style_t = []
            # calculate truncation latent on the fly
            if truncation_latent is None and not hasattr(
                    self, 'truncation_latent'):
                self.truncation_latent = self.get_mean_latent()
                truncation_latent = self.truncation_latent
            elif truncation_latent is None and hasattr(self,
                                                       'truncation_latent'):
                truncation_latent = self.truncation_latent

            for style in styles:
                style_t.append(truncation_latent + truncation *
                               (style - truncation_latent))

            styles = style_t
        # no style mixing
        if len(styles) < 2:
            inject_index = self.num_latents

            if styles[0].ndim < 3:
                latent = styles[0].unsqueeze(1).repeat(1, inject_index, 1)

            else:
                latent = styles[0]
        # style mixing
        else:
            if inject_index is None:
                inject_index = random.randint(1, self.num_latents - 1)

            latent = styles[0].unsqueeze(1).repeat(1, inject_index, 1)
            latent2 = styles[1].unsqueeze(1).repeat(
                1, self.num_latents - inject_index, 1)

            latent = torch.cat([latent, latent2], 1)

        if return_latent_only:
            return latent

        feats = []
        with autocast(enabled=self.fp16_enabled):
            # 4x4 stage
            out = self.constant_input(latent)
            if self.fp16_enabled:
                out = out.to(torch.float16)
            out = self.conv1(
                out,
                latent[:, 0],
                noise=injected_noise[0],
                add_noise=add_noise)
            feats.append(out)
            skip = self.to_rgb1(out, latent[:, 1])

            _index = 1

            # 8x8 ---> higher resolutions
            for up_conv, conv, noise1, noise2, to_rgb in zip(
                    self.convs[::2], self.convs[1::2], injected_noise[1::2],
                    injected_noise[2::2], self.to_rgbs):
                out = up_conv(
                    out, latent[:, _index], noise=noise1, add_noise=add_noise)
                out = conv(
                    out,
                    latent[:, _index + 1],
                    noise=noise2,
                    add_noise=add_noise)
                feats.append(out)
                skip = to_rgb(out, latent[:, _index + 2], skip)
                _index += 2

        # make sure the output image is torch.float32 to avoid RunTime Error
        # in other modules
        img = skip.to(torch.float32)
        if self.bgr2rgb:
            img = torch.flip(img, dims=[1])

        if return_latents or return_noise:
            if return_features:
                output_dict = dict(
                    fake_img=img,
                    latent=latent,
                    inject_index=inject_index,
                    noise_batch=noise_batch,
                    feats=feats[feat_idx])
                return output_dict

            output_dict = dict(
                fake_img=img,
                latent=latent,
                inject_index=inject_index,
                noise_batch=noise_batch)
            return output_dict

        return img
